# Copyright 2024 Volvo Car Corporation
# Licensed under Apache 2.0.

# -*- coding: utf-8 -*-
"""Module for creating a project a2l-file from a template."""

from string import Template


class A2lProjectTemplate:
    """Class for A2l Project Template."""
    def __init__(self, spm_a2l, asap2_version, project_name, events, ip_address, ip_port):
        self.spm_a2l = spm_a2l
        self.asap2_version = asap2_version
        self.project_name = project_name
        self.events = events
        self.ip_address = ip_address
        self.ip_port = ip_port
        self.event_template = Template(
            '        /begin EVENT\n'
            '          "$name" /* EVENT CHANNEL NAME */\n'
            '          "$short_name" /* EVENT CHANNEL SHORT NAME */\n'
            '          $channel_id /* EVENT CHANNEL NUMBER */\n'
            '          DAQ_STIM\n'
            '          0xFF\n'
            '          $time_cycle /* EVENT TIME CYCLE */\n'
            '          $time_unit /* EVENT TIME UNIT */\n'
            '          0x00\n'
            '        /end EVENT\n'
        )
        self.template = Template(
            'ASAP2_VERSION $asap2_version\n'
            '/begin PROJECT $project_name ""\n'
            '  /begin MODULE XCP ""\n'
            '    /begin A2ML\n'
            '      struct Protocol_Layer {\n'
            '        uint;  /* XCP protocol layer version, current 0x100 */\n'
            '        uint;  /* T1 [ms] */\n'
            '        uint;  /* T2 [ms] */\n'
            '        uint;  /* T3 [ms] */\n'
            '        uint;  /* T4 [ms] */\n'
            '        uint;  /* T5 [ms] */\n'
            '        uint;  /* T6 [ms] */\n'
            '        uint;  /* T7 [ms] */\n'
            '        uchar;  /* MAX_CTO */\n'
            '        uint;  /* MAX_DTO */\n'
            '        enum {\n'
            '          "BYTE_ORDER_MSB_LAST" = 0,\n'
            '          "BYTE_ORDER_MSB_FIRST" = 1\n'
            '        };\n'
            '        enum {\n'
            '          "ADDRESS_GRANULARITY_BYTE" = 1,\n'
            '          "ADDRESS_GRANULARITY_WORD" = 2,\n'
            '          "ADDRESS_GRANULARITY_DWORD" = 4\n'
            '        };\n'
            '        taggedstruct {\n'
            '          ("OPTIONAL_CMD" enum {\n'
            '            "GET_COMM_MODE_INFO" = 251,\n'
            '            "GET_ID" = 250,\n'
            '            "SET_REQUEST" = 249,\n'
            '            "GET_SEED" = 248,\n'
            '            "UNLOCK" = 247,\n'
            '            "SET_MTA" = 246,\n'
            '            "UPLOAD" = 245,\n'
            '            "SHORT_UPLOAD" = 244,\n'
            '            "BUILD_CHECKSUM" = 243,\n'
            '            "TRANSPORT_LAYER_CMD" = 242,\n'
            '            "USER_CMD" = 241,\n'
            '            "DOWNLOAD" = 240,\n'
            '            "DOWNLOAD_NEXT" = 239,\n'
            '            "DOWNLOAD_MAX" = 238,\n'
            '            "SHORT_DOWNLOAD" = 237,\n'
            '            "MODIFY_BITS" = 236,\n'
            '            "SET_CAL_PAGE" = 235,\n'
            '            "GET_CAL_PAGE" = 234,\n'
            '            "GET_PAG_PROCESSOR_INFO" = 233,\n'
            '            "GET_SEGMENT_INFO" = 232,\n'
            '            "GET_PAGE_INFO" = 231,\n'
            '            "SET_SEGMENT_MODE" = 230,\n'
            '            "GET_SEGMENT_MODE" = 229,\n'
            '            "COPY_CAL_PAGE" = 228,\n'
            '            "CLEAR_DAQ_LIST" = 227,\n'
            '            "SET_DAQ_PTR" = 226,\n'
            '            "WRITE_DAQ" = 225,\n'
            '            "SET_DAQ_LIST_MODE" = 224,\n'
            '            "GET_DAQ_LIST_MODE" = 223,\n'
            '            "START_STOP_DAQ_LIST" = 222,\n'
            '            "START_STOP_SYNCH" = 221,\n'
            '            "GET_DAQ_CLOCK" = 220,\n'
            '            "READ_DAQ" = 219,\n'
            '            "GET_DAQ_PROCESSOR_INFO" = 218,\n'
            '            "GET_DAQ_RESOLUTION_INFO" = 217,\n'
            '            "GET_DAQ_LIST_INFO" = 216,\n'
            '            "GET_DAQ_EVENT_INFO" = 215,\n'
            '            "FREE_DAQ" = 214,\n'
            '            "ALLOC_DAQ" = 213,\n'
            '            "ALLOC_ODT" = 212,\n'
            '            "ALLOC_ODT_ENTRY" = 211,\n'
            '            "PROGRAM_START" = 210,\n'
            '            "PROGRAM_CLEAR" = 209,\n'
            '            "PROGRAM" = 208,\n'
            '            "PROGRAM_RESET" = 207,\n'
            '            "GET_PGM_PROCESSOR_INFO" = 206,\n'
            '            "GET_SECTOR_INFO" = 205,\n'
            '            "PROGRAM_PREPARE" = 204,\n'
            '            "PROGRAM_FORMAT" = 203,\n'
            '            "PROGRAM_NEXT" = 202,\n'
            '            "PROGRAM_MAX" = 201,\n'
            '            "PROGRAM_VERIFY" = 200\n'
            '          })*;\n'
            '          "COMMUNICATION_MODE_SUPPORTED" taggedunion {\n'
            '            "BLOCK" taggedstruct {\n'
            '              "SLAVE" ;\n'
            '              "MASTER" struct {\n'
            '                uchar;  /* MAX_BS */\n'
            '                uchar;  /* MIN_ST */\n'
            '              };\n'
            '            };\n'
            '            "INTERLEAVED" uchar;  /* QUEUE_SIZE */\n'
            '          };\n'
            '          "SEED_AND_KEY_EXTERNAL_FUNCTION" char[256];/* Name of the Seed&Key function */\n'
            '        };\n'
            '      };\n'
            '\n'
            '      struct Daq {\n'
            '        enum {\n'
            '          "STATIC" = 0,\n'
            '          "DYNAMIC" = 1\n'
            '        };\n'
            '        uint;  /* MAX_DAQ */\n'
            '        uint;  /* MAX_EVENT_CHANNEL */\n'
            '        uchar;  /* MIN_DAQ */\n'
            '        enum {\n'
            '          "OPTIMISATION_TYPE_DEFAULT" = 0,\n'
            '          "OPTIMISATION_TYPE_ODT_TYPE_16" = 1,\n'
            '          "OPTIMISATION_TYPE_ODT_TYPE_32" = 2,\n'
            '          "OPTIMISATION_TYPE_ODT_TYPE_64" = 3,\n'
            '          "OPTIMISATION_TYPE_ODT_TYPE_ALIGNMENT" = 4,\n'
            '          "OPTIMISATION_TYPE_MAX_ENTRY_SIZE" = 5\n'
            '        };\n'
            '        enum {\n'
            '          "ADDRESS_EXTENSION_FREE" = 0,\n'
            '          "ADDRESS_EXTENSION_ODT" = 1,\n'
            '          "ADDRESS_EXTENSION_DAQ" = 3\n'
            '        };\n'
            '        enum {\n'
            '          "IDENTIFICATION_FIELD_TYPE_ABSOLUTE" = 0,\n'
            '          "IDENTIFICATION_FIELD_TYPE_RELATIVE_BYTE" = 1,\n'
            '          "IDENTIFICATION_FIELD_TYPE_RELATIVE_WORD" = 2,\n'
            '          "IDENTIFICATION_FIELD_TYPE_RELATIVE_WORD_ALIGNED" = 3\n'
            '        };\n'
            '        enum {\n'
            '          "GRANULARITY_ODT_ENTRY_SIZE_DAQ_BYTE" = 1,\n'
            '          "GRANULARITY_ODT_ENTRY_SIZE_DAQ_WORD" = 2,\n'
            '          "GRANULARITY_ODT_ENTRY_SIZE_DAQ_DWORD" = 4,\n'
            '          "GRANULARITY_ODT_ENTRY_SIZE_DAQ_DLONG" = 8\n'
            '        };\n'
            '        uchar;  /* MAX_ODT_ENTRY_SIZE_DAQ */\n'
            '        enum {\n'
            '          "NO_OVERLOAD_INDICATION" = 0,\n'
            '          "OVERLOAD_INDICATION_PID" = 1,\n'
            '          "OVERLOAD_INDICATION_EVENT" = 2\n'
            '        };\n'
            '        taggedstruct {\n'
            '          "PRESCALER_SUPPORTED" ;\n'
            '          "RESUME_SUPPORTED" ;\n'
            '          block "STIM" struct {\n'
            '            enum {\n'
            '              "GRANULARITY_ODT_ENTRY_SIZE_STIM_BYTE" = 1,\n'
            '              "GRANULARITY_ODT_ENTRY_SIZE_STIM_WORD" = 2,\n'
            '              "GRANULARITY_ODT_ENTRY_SIZE_STIM_DWORD" = 4,\n'
            '              "GRANULARITY_ODT_ENTRY_SIZE_STIM_DLONG" = 8\n'
            '            };\n'
            '            uchar;  /* MAX_ODT_ENTRY_SIZE_STIM */\n'
            '            taggedstruct {\n'
            '              "BIT_STIM_SUPPORTED" ;\n'
            '            };\n'
            '          };\n'
            '          block "TIMESTAMP_SUPPORTED" struct {\n'
            '            uint;  /* TIMESTAMP_TICKS */\n'
            '            enum {\n'
            '              "NO_TIME_STAMP" = 0,\n'
            '              "SIZE_BYTE" = 1,\n'
            '              "SIZE_WORD" = 2,\n'
            '              "SIZE_DWORD" = 4\n'
            '            };\n'
            '            enum {\n'
            '              "UNIT_1NS" = 0,\n'
            '              "UNIT_10NS" = 1,\n'
            '              "UNIT_100NS" = 2,\n'
            '              "UNIT_1US" = 3,\n'
            '              "UNIT_10US" = 4,\n'
            '              "UNIT_100US" = 5,\n'
            '              "UNIT_1MS" = 6,\n'
            '              "UNIT_10MS" = 7,\n'
            '              "UNIT_100MS" = 8,\n'
            '              "UNIT_1S" = 9\n'
            '            };\n'
            '            taggedstruct {\n'
            '              "TIMESTAMP_FIXED" ;\n'
            '            };\n'
            '          };\n'
            '          "PID_OFF_SUPPORTED" ;\n'
            '          (block "DAQ_LIST" struct {\n'
            '            uint;  /* DAQ_LIST_NUMBER */\n'
            '            taggedstruct {\n'
            '              "DAQ_LIST_TYPE" enum {\n'
            '                "DAQ" = 1,\n'
            '                "STIM" = 2,\n'
            '                "DAQ_STIM" = 3\n'
            '              };\n'
            '              "MAX_ODT" uchar;\n'
            '              "MAX_ODT_ENTRIES" uchar;\n'
            '              "FIRST_PID" uchar;\n'
            '              "EVENT_FIXED" uint;\n'
            '              block "PREDEFINED" taggedstruct {\n'
            '                (block "ODT" struct {\n'
            '                  uchar;  /* ODT number */\n'
            '                  taggedstruct {\n'
            '                    ("ODT_ENTRY" struct {\n'
            '                      uchar;  /* ODT_ENTRY number */\n'
            '                      ulong;  /* address of element */\n'
            '                      uchar;  /* address extension of element */\n'
            '                      uchar;  /* size of element [AG] */\n'
            '                      uchar;  /* BIT_OFFSET */\n'
            '                    })*;\n'
            '                  };\n'
            '                })*;\n'
            '              };\n'
            '            };\n'
            '          })*;\n'
            '          (block "EVENT" struct {\n'
            '            char[101];  /* EVENT_CHANNEL_NAME */\n'
            '            char[9];  /* EVENT_CHANNEL_SHORT_NAME */\n'
            '            uint;  /* EVENT_CHANNEL_NUMBER */\n'
            '            enum {\n'
            '              "DAQ" = 1,\n'
            '              "STIM" = 2,\n'
            '              "DAQ_STIM" = 3\n'
            '            };\n'
            '            uchar;  /* MAX_DAQ_LIST */\n'
            '            uchar;  /* TIME_CYCLE   */\n'
            '            uchar;  /* TIME_UNIT    */\n'
            '            uchar;  /* PRIORITY     */\n'
            '          })*;\n'
            '        };\n'
            '      };\n'
            '\n'
            '      taggedunion Daq_Event {\n'
            '        "FIXED_EVENT_LIST" taggedstruct {\n'
            '          ("EVENT" uint)*;\n'
            '        };\n'
            '        "VARIABLE" taggedstruct {\n'
            '          block "AVAILABLE_EVENT_LIST" taggedstruct {\n'
            '            ("EVENT" uint)*;\n'
            '          };\n'
            '          block "DEFAULT_EVENT_LIST" taggedstruct {\n'
            '            ("EVENT" uint)*;\n'
            '          };\n'
            '        };\n'
            '      };\n'
            '\n'
            '      struct Pag {\n'
            '        uchar;  /* MAX_SEGMENTS */\n'
            '        taggedstruct {\n'
            '          "FREEZE_SUPPORTED" ;\n'
            '        };\n'
            '      };\n'
            '\n'
            '      struct Pgm {\n'
            '        enum {\n'
            '          "PGM_MODE_ABSOLUTE" = 1,\n'
            '          "PGM_MODE_FUNCTIONAL" = 2,\n'
            '          "PGM_MODE_ABSOLUTE_AND_FUNCTIONAL" = 3\n'
            '        };\n'
            '        uchar;  /* MAX_SECTORS */\n'
            '        uchar;  /* MAX_CTO_PGM */\n'
            '        taggedstruct {\n'
            '          (block "SECTOR" struct {\n'
            '            char[101];  /* SECTOR_NAME */\n'
            '            uchar;  /* SECTOR_NUMBER */\n'
            '            ulong;  /* Address */\n'
            '            ulong;  /* Length  */\n'
            '            uchar;  /* CLEAR_SEQUENCE_NUMBER */\n'
            '            uchar;  /* PROGRAM_SEQUENCE_NUMBER */\n'
            '            uchar;  /* PROGRAM_METHOD */\n'
            '          })*;\n'
            '          "COMMUNICATION_MODE_SUPPORTED" taggedunion {\n'
            '            "BLOCK" taggedstruct {\n'
            '              "SLAVE" ;\n'
            '              "MASTER" struct {\n'
            '                uchar;  /* MAX_BS_PGM */\n'
            '                uchar;  /* MIN_ST_PGM */\n'
            '              };\n'
            '            };\n'
            '            "INTERLEAVED" uchar;  /* QUEUE_SIZE_PGM */\n'
            '          };\n'
            '        };\n'
            '      };\n'
            '\n'
            '      struct Segment {\n'
            '        uchar;  /* SEGMENT_NUMBER */\n'
            '        uchar;  /* number of pages */\n'
            '        uchar;  /* ADDRESS_EXTENSION */\n'
            '        uchar;  /* COMPRESSION_METHOD */\n'
            '        uchar;  /* ENCRYPTION_METHOD */\n'
            '        taggedstruct {\n'
            '          block "CHECKSUM" struct {\n'
            '            enum {\n'
            '              "XCP_ADD_11" = 1,\n'
            '              "XCP_ADD_12" = 2,\n'
            '              "XCP_ADD_14" = 3,\n'
            '              "XCP_ADD_22" = 4,\n'
            '              "XCP_ADD_24" = 5,\n'
            '              "XCP_ADD_44" = 6,\n'
            '              "XCP_CRC_16" = 7,\n'
            '              "XCP_CRC_16_CITT" = 8,\n'
            '              "XCP_CRC_32" = 9,\n'
            '              "XCP_USER_DEFINED" = 255\n'
            '            };\n'
            '            taggedstruct {\n'
            '              "MAX_BLOCK_SIZE" ulong;\n'
            '              "EXTERNAL_FUNCTION" char[256];  /* Name of the Checksum.DLL */\n'
            '            };\n'
            '          };\n'
            '          (block "PAGE" struct {\n'
            '            uchar;  /* PAGE_NUMBER */\n'
            '            enum {\n'
            '              "ECU_ACCESS_NOT_ALLOWED" = 0,\n'
            '              "ECU_ACCESS_WITHOUT_XCP_ONLY" = 1,\n'
            '              "ECU_ACCESS_WITH_XCP_ONLY" = 2,\n'
            '              "ECU_ACCESS_DONT_CARE" = 3\n'
            '            };\n'
            '            enum {\n'
            '              "XCP_READ_ACCESS_NOT_ALLOWED" = 0,\n'
            '              "XCP_READ_ACCESS_WITHOUT_ECU_ONLY" = 1,\n'
            '              "XCP_READ_ACCESS_WITH_ECU_ONLY" = 2,\n'
            '              "XCP_READ_ACCESS_DONT_CARE" = 3\n'
            '            };\n'
            '            enum {\n'
            '              "XCP_WRITE_ACCESS_NOT_ALLOWED" = 0,\n'
            '              "XCP_WRITE_ACCESS_WITHOUT_ECU_ONLY" = 1,\n'
            '              "XCP_WRITE_ACCESS_WITH_ECU_ONLY" = 2,\n'
            '              "XCP_WRITE_ACCESS_DONT_CARE" = 3\n'
            '            };\n'
            '            taggedstruct {\n'
            '              "INIT_SEGMENT" uchar;  /* references segment that initialises this page */\n'
            '            };\n'
            '          })*;\n'
            '          (block "ADDRESS_MAPPING" struct {\n'
            '            ulong;  /* source address */\n'
            '            ulong;  /* destination address */\n'
            '            ulong;  /* length */\n'
            '          })*;\n'
            '          "PGM_VERIFY" ulong;  /* verification value for PGM */\n'
            '        };\n'
            '      };\n'
            '\n'
            '      taggedstruct Common_Parameters {\n'
            '        block "PROTOCOL_LAYER" struct Protocol_Layer;\n'
            '        block "SEGMENT" struct Segment;\n'
            '        block "DAQ" struct Daq;\n'
            '        block "PAG" struct Pag;\n'
            '        block "PGM" struct Pgm;\n'
            '        block "DAQ_EVENT" taggedunion Daq_Event;\n'
            '      };\n'
            '\n'
            '      struct UDP_IP_Parameters {\n'
            '        uint;  /* XCP on UDP_IP version, currently 0x0100 */\n'
            '        uint;  /* PORT */\n'
            '        taggedunion {\n'
            '          "HOST_NAME" char[256];\n'
            '          "ADDRESS" char[15];\n'
            '        };\n'
            '      };\n'
            '\n'
            '      block "IF_DATA" taggedunion if_data {\n'
            '        "XCP" struct {\n'
            '          taggedstruct Common_Parameters;  /* default parameters */\n'
            '          taggedstruct {\n'
            '            block "XCP_ON_UDP_IP" struct {\n'
            '              struct UDP_IP_Parameters;  /* specific for UDP_IP */\n'
            '              taggedstruct Common_Parameters;  /* overruling of default */\n'
            '            };\n'
            '          };\n'
            '        };\n'
            '      };\n'
            '    /end A2ML\n'
            '\n'
            '    /begin MOD_COMMON ""\n'
            '      BYTE_ORDER MSB_LAST\n'
            '      ALIGNMENT_BYTE 1\n'
            '      ALIGNMENT_WORD 1\n'
            '      ALIGNMENT_LONG 1\n'
            '      ALIGNMENT_FLOAT32_IEEE 1\n'
            '      ALIGNMENT_FLOAT64_IEEE 1\n'
            '    /end MOD_COMMON\n'
            '\n'
            '    /begin IF_DATA XCP\n'
            '      /begin PROTOCOL_LAYER\n'
            '        0x0100\n'
            '        0x03E8\n'
            '        0xC8\n'
            '        0x00\n'
            '        0x00\n'
            '        0x00\n'
            '        0x00\n'
            '        0x00\n'
            '        0x60\n'
            '        0x12C\n'
            '        BYTE_ORDER_MSB_LAST\n'
            '        ADDRESS_GRANULARITY_BYTE\n'
            '        OPTIONAL_CMD ALLOC_ODT_ENTRY\n'
            '        OPTIONAL_CMD ALLOC_ODT\n'
            '        OPTIONAL_CMD ALLOC_DAQ\n'
            '        OPTIONAL_CMD FREE_DAQ\n'
            '        OPTIONAL_CMD GET_DAQ_RESOLUTION_INFO\n'
            '        OPTIONAL_CMD GET_DAQ_PROCESSOR_INFO\n'
            '        OPTIONAL_CMD START_STOP_SYNCH\n'
            '        OPTIONAL_CMD GET_DAQ_CLOCK\n'
            '        OPTIONAL_CMD START_STOP_DAQ_LIST\n'
            '        OPTIONAL_CMD GET_DAQ_LIST_MODE\n'
            '        OPTIONAL_CMD SET_DAQ_LIST_MODE\n'
            '        OPTIONAL_CMD WRITE_DAQ\n'
            '        OPTIONAL_CMD SET_DAQ_PTR\n'
            '        OPTIONAL_CMD CLEAR_DAQ_LIST\n'
            '        OPTIONAL_CMD SHORT_UPLOAD\n'
            '        OPTIONAL_CMD UPLOAD\n'
            '        OPTIONAL_CMD SET_MTA\n'
            '        OPTIONAL_CMD GET_ID\n'
            '        OPTIONAL_CMD GET_COMM_MODE_INFO\n'
            '        OPTIONAL_CMD BUILD_CHECKSUM\n'
            '        OPTIONAL_CMD SET_SEGMENT_MODE\n'
            '        OPTIONAL_CMD GET_SEGMENT_MODE\n'
            '        OPTIONAL_CMD SET_REQUEST\n'
            '        OPTIONAL_CMD GET_SEED\n'
            '        OPTIONAL_CMD UNLOCK\n'
            '        OPTIONAL_CMD COPY_CAL_PAGE\n'
            '        SEED_AND_KEY_EXTERNAL_FUNCTION Seed_Key.dll'
            '/* Add name of seed & key dll file here.*/\n'
            '      /end PROTOCOL_LAYER\n'
            '\n'
            '      /begin DAQ\n'
            '        DYNAMIC\n'
            '        0x00\n'
            '        0x06\n'
            '        0x00\n'
            '        OPTIMISATION_TYPE_DEFAULT\n'
            '        ADDRESS_EXTENSION_FREE\n'
            '        IDENTIFICATION_FIELD_TYPE_ABSOLUTE\n'
            '        GRANULARITY_ODT_ENTRY_SIZE_DAQ_BYTE\n'
            '        0x60\n'
            '        NO_OVERLOAD_INDICATION\n'
            '        /begin TIMESTAMP_SUPPORTED\n'
            '          0x01\n'
            '          SIZE_DWORD\n'
            '          UNIT_100US\n'
            '          TIMESTAMP_FIXED\n'
            '        /end TIMESTAMP_SUPPORTED\n'
            '\n'
            '$events\n'
            '      /end DAQ\n'
            '\n'
            '      /begin PAG\n'
            '        0x01\n'
            '        FREEZE_SUPPORTED\n'
            '      /end PAG\n'
            '      /begin PGM\n'
            '        PGM_MODE_ABSOLUTE\n'
            '        0x01\n'
            '        0x00\n'
            '        /begin SECTOR\n'
            '          "Sector"\n'
            '          0x00\n'
            '          0x22014\n'
            '          0x38\n'
            '          0x00\n'
            '          0x00\n'
            '          0x00\n'
            '        /end SECTOR\n'
            '      /end PGM\n'
            '\n'
            '      /begin XCP_ON_UDP_IP\n'
            '        0x0100\n'
            '        $ip_port\n'
            '        ADDRESS "$ip_address"\n'
            '      /end XCP_ON_UDP_IP\n'
            '    /end IF_DATA\n'
            '\n'
            '    /begin MOD_PAR ""\n'
            '      /begin MEMORY_SEGMENT\n'
            '        caldata "caldata"\n'
            '        DATA FLASH INTERN 0x00000 0x0000 -1 -1 -1 -1 -1\n'
            '        /begin IF_DATA XCP\n'
            '          /begin SEGMENT\n'
            '            0x00\n'
            '            0x02\n'
            '            0x00\n'
            '            0x00\n'
            '            0x00\n'
            '            /begin CHECKSUM\n'
            '              XCP_CRC_16_CITT\n'
            '            /end CHECKSUM\n'
            '            /begin PAGE\n'
            '              0x00\n'
            '              ECU_ACCESS_WITH_XCP_ONLY\n'
            '              XCP_READ_ACCESS_WITH_ECU_ONLY\n'
            '              XCP_WRITE_ACCESS_NOT_ALLOWED\n'
            '              INIT_SEGMENT 0x00\n'
            '            /end PAGE\n'
            '            /begin PAGE\n'
            '              0x01\n'
            '              ECU_ACCESS_WITH_XCP_ONLY\n'
            '              XCP_READ_ACCESS_WITH_ECU_ONLY\n'
            '              XCP_WRITE_ACCESS_WITH_ECU_ONLY\n'
            '              INIT_SEGMENT 0x00\n'
            '            /end PAGE\n'
            '          /end SEGMENT\n'
            '        /end IF_DATA\n'
            '      /end MEMORY_SEGMENT\n'
            '      /begin MEMORY_SEGMENT\n'
            '        text ".text"\n'
            '        CODE ROM INTERN 0x26030 0x72c10 -1 -1 -1 -1 -1\n'
            '        /begin IF_DATA XCP\n'
            '          /begin SEGMENT\n'
            '            0x01\n'
            '            0x01\n'
            '            0x00\n'
            '            0x00\n'
            '            0x00\n'
            '            /begin CHECKSUM\n'
            '              XCP_CRC_16_CITT\n'
            '            /end CHECKSUM\n'
            '            /begin PAGE\n'
            '              0x01\n'
            '              ECU_ACCESS_WITH_XCP_ONLY\n'
            '              XCP_READ_ACCESS_WITH_ECU_ONLY\n'
            '              XCP_WRITE_ACCESS_WITH_ECU_ONLY\n'
            '              INIT_SEGMENT 0x00\n'
            '            /end PAGE\n'
            '          /end SEGMENT\n'
            '        /end IF_DATA\n'
            '      /end MEMORY_SEGMENT\n'
            '    /end MOD_PAR\n'
            '\n'
            '    /begin COMPU_METHOD BitSlice.CONVERSION ""\n'
            '      RAT_FUNC "%6.2f" ""\n'
            '      COEFFS 0 1 0 0 0 1\n'
            '    /end COMPU_METHOD\n'
            '\n'
            '    /* SPM declarations start. */\n'
            '\n'
            '$spm_a2l\n'
            '\n'
            '    /* SPM declarations end. */\n'
            '\n'
            '    /begin RECORD_LAYOUT __UBYTE_Z\n'
            '      FNC_VALUES 1 UBYTE ROW_DIR DIRECT\n'
            '    /end RECORD_LAYOUT\n'
            '\n'
            '    /begin RECORD_LAYOUT __UWORD_Z\n'
            '      FNC_VALUES 1 UWORD ROW_DIR DIRECT\n'
            '    /end RECORD_LAYOUT\n'
            '\n'
            '    /begin RECORD_LAYOUT __ULONG_Z\n'
            '      FNC_VALUES 1 ULONG ROW_DIR DIRECT\n'
            '    /end RECORD_LAYOUT\n'
            '\n'
            '    /begin RECORD_LAYOUT __SBYTE_Z\n'
            '      FNC_VALUES 1 SBYTE ROW_DIR DIRECT\n'
            '    /end RECORD_LAYOUT\n'
            '\n'
            '    /begin RECORD_LAYOUT __SWORD_Z\n'
            '      FNC_VALUES 1 SWORD ROW_DIR DIRECT\n'
            '    /end RECORD_LAYOUT\n'
            '\n'
            '    /begin RECORD_LAYOUT __SLONG_Z\n'
            '      FNC_VALUES 1 SLONG ROW_DIR DIRECT\n'
            '    /end RECORD_LAYOUT\n'
            '\n'
            '    /begin RECORD_LAYOUT __FLOAT32_IEEE_Z\n'
            '      FNC_VALUES 1 FLOAT32_IEEE ROW_DIR DIRECT\n'
            '    /end RECORD_LAYOUT\n'
            '\n'
            '    /begin RECORD_LAYOUT __FLOAT64_IEEE_Z\n'
            '      FNC_VALUES 1 FLOAT64_IEEE ROW_DIR DIRECT\n'
            '    /end RECORD_LAYOUT\n'
            '\n'
            '    /begin RECORD_LAYOUT __UBYTE_S\n'
            '      FNC_VALUES 1 UBYTE COLUMN_DIR DIRECT\n'
            '    /end RECORD_LAYOUT\n'
            '\n'
            '    /begin RECORD_LAYOUT __UWORD_S\n'
            '      FNC_VALUES 1 UWORD COLUMN_DIR DIRECT\n'
            '    /end RECORD_LAYOUT\n'
            '\n'
            '    /begin RECORD_LAYOUT __ULONG_S\n'
            '      FNC_VALUES 1 ULONG COLUMN_DIR DIRECT\n'
            '    /end RECORD_LAYOUT\n'
            '\n'
            '    /begin RECORD_LAYOUT __SBYTE_S\n'
            '      FNC_VALUES 1 SBYTE COLUMN_DIR DIRECT\n'
            '    /end RECORD_LAYOUT\n'
            '\n'
            '    /begin RECORD_LAYOUT __SWORD_S\n'
            '      FNC_VALUES 1 SWORD COLUMN_DIR DIRECT\n'
            '    /end RECORD_LAYOUT\n'
            '\n'
            '    /begin RECORD_LAYOUT __SLONG_S\n'
            '      FNC_VALUES 1 SLONG COLUMN_DIR DIRECT\n'
            '    /end RECORD_LAYOUT\n'
            '\n'
            '    /begin RECORD_LAYOUT __FLOAT32_IEEE_S\n'
            '      FNC_VALUES 1 FLOAT32_IEEE COLUMN_DIR DIRECT\n'
            '    /end RECORD_LAYOUT\n'
            '\n'
            '    /begin RECORD_LAYOUT __FLOAT64_IEEE_S\n'
            '      FNC_VALUES 1 FLOAT64_IEEE COLUMN_DIR DIRECT\n'
            '    /end RECORD_LAYOUT\n'
            '\n'
            '    /begin RECORD_LAYOUT SSV__UBYTE_S\n'
            '      AXIS_PTS_X 1 UBYTE INDEX_INCR DIRECT\n'
            '    /end RECORD_LAYOUT\n'
            '\n'
            '    /begin RECORD_LAYOUT SSV__UWORD_S\n'
            '      AXIS_PTS_X 1 UWORD INDEX_INCR DIRECT\n'
            '    /end RECORD_LAYOUT\n'
            '\n'
            '    /begin RECORD_LAYOUT SSV__ULONG_S\n'
            '      AXIS_PTS_X 1 ULONG INDEX_INCR DIRECT\n'
            '    /end RECORD_LAYOUT\n'
            '\n'
            '    /begin RECORD_LAYOUT SSV__SBYTE_S\n'
            '      AXIS_PTS_X 1 SBYTE INDEX_INCR DIRECT\n'
            '    /end RECORD_LAYOUT\n'
            '\n'
            '    /begin RECORD_LAYOUT SSV__SWORD_S\n'
            '      AXIS_PTS_X 1 SWORD INDEX_INCR DIRECT\n'
            '    /end RECORD_LAYOUT\n'
            '\n'
            '    /begin RECORD_LAYOUT SSV__SLONG_S\n'
            '      AXIS_PTS_X 1 SLONG INDEX_INCR DIRECT\n'
            '    /end RECORD_LAYOUT\n'
            '\n'
            '    /begin RECORD_LAYOUT SSV__FLOAT32_IEEE_S\n'
            '      AXIS_PTS_X 1 FLOAT32_IEEE INDEX_INCR DIRECT\n'
            '    /end RECORD_LAYOUT\n'
            '\n'
            '    /begin RECORD_LAYOUT SSV__FLOAT64_IEEE_S\n'
            '      AXIS_PTS_X 1 FLOAT64_IEEE INDEX_INCR DIRECT\n'
            '    /end RECORD_LAYOUT\n'
            '\n'
            '  /end MODULE\n'
            '/end PROJECT\n'
        )

    def render(self):
        """Render the complete A2L.
        Returns:
            a2l (str): The A2L for the project.
        """
        event_a2l = ''
        for event in self.events:
            event_a2l += self.event_template.substitute(
                name=event['name'],
                short_name=event['name'],
                channel_id=event['channel_id'],
                time_cycle=event['time_cycle'],
                time_unit=event['time_unit']
            )
        return self.template.substitute(
            events=event_a2l,
            asap2_version=self.asap2_version,
            project_name=self.project_name.replace('-', '_'),
            ip_address=self.ip_address,
            ip_port=self.ip_port,
            spm_a2l=self.spm_a2l
        )


class A2lSilverTemplate:
    """Class for A2l Silver Template."""
    def __init__(self, base_a2l):
        """Init."""
        self.patched_a2l = self.basic_patching(base_a2l)
        self.template = Template(
            '/* QTronic Header */\n'
            'ASAP2_VERSION 1 6\n'
            '/begin PROJECT SPM ""\n'
            '/begin MODULE SPM ""\n'
            '\n'
            '$base_a2l'
            '\n'
            '/end MODULE\n'
            '/end PROJECT'
        )

    @staticmethod
    def basic_patching(base_a2l):
        """Perform required basic patching.

        Args:
            base_a2l (str): Inital A2l content.
        Returns:
            (str): Patched A2L content.
        """
        return "".join([char if ord(char) < 128 else "?" for char in base_a2l])

    def render(self):
        """Render the complete A2L.

        Returns:
            (str): The patched A2L for Silver.
        """
        if "ASAP2_VERSION" in self.patched_a2l[:999]:
            return self.patched_a2l
        return self.template.substitute(base_a2l=self.patched_a2l)



# Copyright 2024 Volvo Car Corporation
# Licensed under Apache 2.0.

"""Module for handling the update of model yaml files."""


import argparse
import sys
from pathlib import Path
from typing import List, Optional

from ruamel.yaml import YAML
from powertrain_build.interface.application import Application
from powertrain_build.interface.base import BaseApplication

PARSER_HELP = "Update model yaml files."


class BadYamlFormat(Exception):
    """Exception to raise when signal is not in/out signal."""
    def __init__(self, signal):
        self.message = f"{signal} is not in-signal or out-signal."


class UpdateYmlFormat(BaseApplication):
    """Class to handle the update of model yaml files."""

    def __init__(self, base_application):
        self.base_application = base_application
        self.raw = {}
        self.app_insignals = self.get_insignals_name()
        self.app_outsignals = self.get_outsignals_name()

    def read_translation(self, translation_file):
        """Read specification of the yaml file.

        Args:
            translation_file (Path): file with specs.
        """
        if not translation_file.is_file():
            return {}
        with open(translation_file, encoding="utf-8") as translation:
            yaml = YAML(typ='safe', pure=True)
            self.raw = yaml.load(translation)

    def parse_groups(self, signal_groups):
        """Parse signal groups.

        Args:
            signal_groups (dict): Hal/dp signal group in yaml file.
        """
        for interface_type, group_definitions in signal_groups.items():
            for group in group_definitions:
                for signals in group.values():
                    self.check_signals({interface_type: signals})

    def parse_hal_definitions(self, hal):
        """Parse hal.

        Args:
            hal (dict): hal in yaml file.
        """
        self.parse_groups(hal)

    def parse_service_definitions(self, service):
        """Parse service.

        Args:
            service (dict): service in yaml file.
        """
        for interface, definition in service.items():
            for apis in definition['properties']:
                for signals in apis.values():
                    self.check_signals({interface: signals})

    def check_signals(self, signals_definition):
        """check signal direction(in-signal or out-signal).

        Args:
            signals_definition (dict): parsed yaml file.
        """
        for specifications in signals_definition.values():
            for specification in specifications:
                in_out_signal = [key for key in specification.keys() if 'signal' in key]
                if in_out_signal == []:
                    raise ValueError(f"signal is not defined for property: {specification['property']} !")
                if "in" in in_out_signal[0]:
                    if specification[in_out_signal[0]] not in self.app_insignals:
                        if specification[in_out_signal[0]] in self.app_outsignals:
                            specification["outsignal"] = specification[in_out_signal[0]]
                        else:
                            raise BadYamlFormat(specification[in_out_signal[0]])
                if "out" in in_out_signal[0]:
                    if specification[in_out_signal[0]] not in self.app_outsignals:
                        if specification[in_out_signal[0]] in self.app_insignals:
                            specification["insignal"] = specification[in_out_signal[0]]
                        else:
                            raise BadYamlFormat(specification[in_out_signal[0]])
                if in_out_signal[0] == "signal":
                    if specification["signal"] in self.app_insignals:
                        specification["insignal"] = specification["signal"]
                        del specification["signal"]
                    elif specification["signal"] in self.app_outsignals:
                        specification["outsignal"] = specification["signal"]
                        del specification["signal"]
                    else:
                        raise BadYamlFormat(specification[in_out_signal[0]])

    def get_insignals_name(self):
        """Base application in-signals.
        """
        app_insignals = []
        for signal in self.base_application.insignals:
            app_insignals.append(signal.name)
        return app_insignals

    def get_outsignals_name(self):
        """Base application out-signals.
        """
        app_outsignals = []
        for signal in self.base_application.outsignals:
            app_outsignals.append(signal.name)
        return app_outsignals

    def parse_definition(self, definition):
        """Parses all definition files

        Args:
            definition (list(Path)): Definition files
        """
        for translation in definition:
            self.read_translation(translation)
            self.check_signals(self.raw.get("signals", {}))
            self.parse_groups(self.raw.get("signal_groups", {}))
            self.parse_hal_definitions(self.raw.get("hal", {}))
            self.parse_service_definitions(self.raw.get("service", {}))
            self.write_to_file(translation)

    def write_to_file(self, translation):
        """Write in/out signals to file
        Args:
            spec (dict): Specification
            output (Path): File to write
        """
        with open(translation, "w", encoding="utf-8") as new_file:
            yaml = YAML()
            yaml.dump(self.raw, new_file)


def get_app(config):
    """ Get an app specification for the current project

    Args:
        config (pathlib.Path): Path to the ProjectCfg.json.
    Returns:
        app (Application): powertrain_build project.
    """
    app = Application()
    app.parse_definition(config)
    return app


def update_model_yaml_cli(args: argparse.Namespace):
    """CLI for update model yaml."""
    app = get_app(args.config)
    translation_files = app.get_translation_files()
    uymlf = UpdateYmlFormat(app)
    uymlf.parse_definition(translation_files)


def configure_parser(parser: argparse.ArgumentParser):
    """Configure parser for update model yaml."""
    parser.add_argument("config", help="The SPA2 project config file", type=Path)
    parser.set_defaults(func=update_model_yaml_cli)


def main(argv: Optional[List[str]] = None):
    """Main function for update model yaml."""
    parser = argparse.ArgumentParser(description=PARSER_HELP)
    configure_parser(parser)
    args = parser.parse_args(argv)
    return args.func(args)


if __name__ == "__main__":
    main(sys.argv[1:])



# Copyright 2024 Volvo Car Corporation
# Licensed under Apache 2.0.

# -*- coding: utf-8 -*-
"""Python module used for abstracting an application that should be interfacing others."""
from abc import abstractmethod
from ruamel.yaml import YAML
from powertrain_build.lib import logger

LOGGER = logger.create_logger('base')


def filter_signals(signals, domain):
    """ Take a list of signals and remove all domains belonging to a domain

    If the signal is part of the domain, it is not part of the resulting list

    Arguments:
        signals (list(Signal)): signals to filter
        domain (Domain): domain that the signals should not be part of
    """
    filtered_signals = []
    for signal in signals:
        if signal.name not in domain.signals:
            filtered_signals.append(signal)
    return filtered_signals


class MultipleProducersError(Exception):
    """Error when setting a producer and there already exists one"""
    def __init__(self, signal, old_producer, new_producer):
        """Set error message

        Args:
            signal (Signal): Signal object
            old_producer (BaseApplication): Producer already registered
            new_producer (BaseApplication): Producer attempted to be registered
        """
        super().__init__()
        self.message = (f"{signal.name}:"
                        f" Attempting to set producer {new_producer}"
                        f" when {old_producer} is already set")


class BaseApplication:
    """Base application to build other adapters on"""
    name = str()
    _signals = None
    node = str()  # Used to calculate interface

    read_strategies = {
        "Always",
        "OnChanged",
        "OnUpdated"
    }

    def __repr__(self):
        """String representation for logging and debugging

        Returns:
            repr (string): Name of the application, the number of insignals and outsignals
        """
        return f"<{self.name}" \
               f" insignals:{len(self.insignals)}" \
               f" outsignals:{len(self.outsignals)}>"

    def parse_signals(self):
        """API interface to read all signals in any child object"""
        self._get_signals()

    @property
    def insignals(self):
        """ Insignals to the raster.
        Calculated as all read ports - all written ports

        Returns:
            signals (list): List of Signal objects.
        """
        if self._insignals is None:
            self._get_signals()
        return [self._signals[port] for port in self._insignals - self._outsignals]

    @property
    def outsignals(self):
        """ All outports.
        Since we might consume some of the signals that should also be sent elsewhere,
        we do not remove internally consumed signals.

        Returns:
            signals (list): List of Signal objects.
        """
        if self._outsignals is None:
            self._get_signals()
        return [self._signals[port] for port in self._outsignals]

    @property
    def signals(self):
        """API interface property in any child object

        All cached signals.
        If no cache exists, reads all signals and save to cache.

        Returns:
            signals (list): Signal objects
        """
        if self._signals is None:
            self.parse_signals()
        return self._signals.values()

    @abstractmethod
    def _get_signals(self):
        """Stub to implement in child object"""

    @abstractmethod
    def get_signal_properties(self, signal):
        """Stub to implement in child object

        Ideally, this should be moved to the signal.
        Currently, getting the properties depends on how we read and define the signals.
        """

    @abstractmethod
    def parse_definition(self, definition):
        """Stub for parsing a defintion after the object has been initialized.

        Raises NotImplementedError if called without being implemented.

        Args:
            definition: Definition of the Application. Type depends on the application.
        """
        raise NotImplementedError('This is a stub')


class Signal:
    """Signal object

    The signal should behave the same way independently of where we define it.
    """
    def __repr__(self):
        """String representation for logging and debugging

        Returns:
            repr (string): Name of the application, the number of insignals and outsignals
        """
        return (f"<{self.name} in {self.applications}"
                f" producer:{self.producer}"
                f" consumers:{self.consumers}>")

    def __init__(self, name, application):
        """Define base properties of the signal object

        The application object is used to read properties of a signal.
        TODO: Do this when we define the signal and add properties known in other
              systems when we encounter them.

        Args:
            name (string): Signal name
            application (BaseApplication): Application defining the signal
        """
        self.name = name
        self.applications = {}  # Add applications to a dict to prevent duplicates
        if application is not None:
            self.applications[application.name] = application
        self._consumers = set()
        self._producer = None

    def add_application(self, application):
        """Add an application to find properties from

        Args:
            application (BaseApplication): Application to read properties from
        """
        if application.name in self.applications:
            return
        self.applications[application.name] = application

    @property
    def consumers(self):
        """Get all consumers of a signal

        Returns:
            consumers (set): All consumers of a signal
        """
        if isinstance(self._consumers, set):
            return self._consumers
        return set()

    @consumers.setter
    def consumers(self, consumers):
        """Set consumers of a signal

        If the consumers is a list or set, iterate over each consumer
        Otherwise, add the consumer to the set of consumers

        Args:
            consumers (list/set/string): consumer(s) of a signal
        """
        if isinstance(consumers, (list, set)):
            for consumer in consumers:
                self._consumers.add(consumer)
        else:
            self._consumers.add(consumers)

    @property
    def producer(self):
        """Get the producer of a signal

        Since we have some strange signals with multiple producers,
        such as counters for dep, this returns a set.
        Returns:
            producer (set): Producer(s) of a signal
        """
        if isinstance(self._producer, set):
            return self._producer
        return set()

    @producer.setter
    def producer(self, producer):
        """Set producer of a signal

        Args:
            producer (string/set): Name of the producer
        """
        if isinstance(producer, set):
            self._producer = producer
        else:
            self._producer = {producer}

    def set_producer(self, producer):
        """Set producer of a signal

        If there already is a registered producer of the signal,
        raise MultipleProducersError

        This can be expected and force_producer can be called to override this.
        That must be explicit in each instance.

        Args:
            producer (string): Name of the producer
            application (BaseApplication): Application defining the signal. Optional
        """
        if isinstance(producer, set):
            if self._producer is not None and producer - self._producer:
                raise MultipleProducersError(self, self._producer, producer)
            self.producer = producer
        else:
            if self._producer is not None \
                    and isinstance(producer, str) \
                    and producer not in self._producer:
                raise MultipleProducersError(self, self._producer, producer)
            self.producer = {producer}

    def force_producer(self, producer):
        """Forcefully update add producers of a signal

        This is needed since we have some signals that are written by multiple model
        Args:
            producers (string): Producer of a signal
            application (BaseApplication): Application defining the signal. Optional
        """

        self._producer.add(producer)

    @property
    def properties(self):
        """Properties of a signal

        Currently not homogenized.
        Therefore we read the properties from the application that defined the signal.

        Returns:
            properties (dict): properties of a signal
        """
        properties = {}
        for application in self.applications.values():
            LOGGER.debug('Getting properties for %s from %s', self.name, application.name)
            application_properties = application.get_signal_properties(self)
            LOGGER.debug(application_properties)
            for key, value in application_properties.items():
                LOGGER.debug('Looking at %s: %s', key, value)
                if key in properties and value != properties[key]:
                    LOGGER.debug('Signal %s already has %s with value %s, ignoring %s from %s',
                                 self.name, key, properties[key], value, application.name)
                    continue
                properties[key] = value
        return properties


class Interface:
    """Interface between two objects"""
    def __repr__(self):
        """String representation for logging and debugging

        Returns:
            repr (string): Name of the interface, and the length of received and transmitted signals
        """
        return (f"<{self.name}"
                f" a->b:{len(self.get_directional_signals(self.current, self.corresponding))}"
                f" b->a:{len(self.get_directional_signals(self.corresponding, self.current))}>")

    def debug(self):
        """Debug an interface object to stdout"""
        LOGGER.info('name: %s', self.name)
        for signal in self.get_directional_signals(self.current, self.corresponding):
            LOGGER.info('insignal: %s', signal)
        for signal in self.get_directional_signals(self.corresponding, self.current):
            LOGGER.info('outsignal: %s', signal)

    def __init__(self, current, corresponding):
        """Create the interface object

        Args:
            current (BaseApplication): Primary object of an interface
            corresponding (BaseApplication): Secondary object of an interface
        """
        self.name = current.name + '_' + corresponding.name
        self.current = current
        self.corresponding = corresponding

    @staticmethod
    def get_directional_signals(producer, consumer):
        """Get signals going from producer to consumer

        Args:
            producer (BaseApplication): producer of the signals
            consumer (BaseApplication): consumer of the signals
        Returns:
            signals (list): Signals sent from producer and received in consumer
        """
        outsignals = {signal.name: signal for signal in producer.outsignals}
        signals = []
        for signal in consumer.insignals:
            if signal.name in outsignals:
                signal.set_producer(outsignals[signal.name].producer)
                signal.add_application(producer)
                signal.add_application(consumer)
                signals.append(signal)
        return signals

    def get_produced_signals(self, producer_name):
        """Get signals going from producer to consumer

        This function can be used if you are lacking some objects

        Args:
            consumer_name (string): name of the consumer of the signals
        Returns:
            signals (list): Signals sent from producer and received in consumer
        """
        if producer_name == self.current.name:
            consumer = self.corresponding
            producer = self.current
        elif producer_name == self.corresponding.name:
            consumer = self.current
            producer = self.corresponding
        else:
            LOGGER.error('%s not in [%s, %s]',
                         producer_name,
                         self.current.name,
                         self.corresponding.name)
        return self.get_directional_signals(producer, consumer)

    def get_consumed_signals(self, consumer_name):
        """Get signals going from producer to consumer

        This function can be used if you are lacking some objects

        Args:
            consumer_name (string): name of the consumer of the signals
        Returns:
            signals (list): Signals sent from producer and received in consumer
        """
        if consumer_name == self.current.name:
            consumer = self.current
            producer = self.corresponding
        elif consumer_name == self.corresponding.name:
            consumer = self.corresponding
            producer = self.current
        else:
            LOGGER.error('%s not in [%s, %s]',
                         consumer_name,
                         self.current.name,
                         self.corresponding.name)
        return self.get_directional_signals(producer, consumer)


class Domain:
    """Domain with interacting interfaces"""
    def __repr__(self):
        """String representation for logging and debugging

        Returns:
            repr (string): Name of the domain, and all clients for that domain
        """
        return f"<{self.name}: {self.clients}>"

    def __init__(self):
        """Initialize the object"""
        self.name = ''
        self.signals = {}
        self.clients = set()
        self._clients = {}

    def set_name(self, name):
        """Set the name of the domain

        Args:
            name (string): Name of the domain
        """
        self.name = name

    def add_interface(self, interface):
        """Add an interface to a domain

        Args:
            interface (Interface): Interface object
        """
        self._process_interface(interface)

    def _process_interface(self, interface):
        """Process interface to add signals to the domain

        Args:
            interface (Interface): Interface object
        """
        for signal in interface.get_directional_signals(interface.current, interface.corresponding):
            self._process_signal(signal.name, interface.current, interface.corresponding)
        for signal in interface.get_directional_signals(interface.corresponding, interface.current):
            self._process_signal(signal.name, interface.corresponding, interface.current)

    def _process_signal(self, signal_name, producer, consumer):
        """Process signal to add to the domain

        Args:
            signal_name (string): Name of the signal
            consumer (BaseApplication): Consumer application of the signal
            producer (BaseApplication): Producer application of the signal
        """
        if signal_name not in self.signals:
            self.signals[signal_name] = Signal(signal_name, producer)
            self.signals[signal_name].consumers = consumer.name
            self.signals[signal_name].add_application(producer)
            self.signals[signal_name].add_application(consumer)
        signal = self.signals[signal_name]
        if producer.name not in self.clients:
            self.clients.add(producer.name)
            self._clients[producer.name] = {'producer': [], 'consumer': []}
        self._clients[producer.name]['producer'].append(signal)
        if consumer.name not in self.clients:
            self.clients.add(producer.name)
            self._clients[producer.name] = {'producer': [], 'consumer': []}
        self._clients[producer.name]['consumer'].append(signal)
        signal.consumers = consumer.name
        try:
            signal.producer = producer.name
        except MultipleProducersError as mpe:
            LOGGER.debug(mpe.message)

    def create_groups(self):
        """Create groups of signals going from each producer

        Returns:
            signal_groups (dict): Signal groups
        """
        signal_groups = {}
        for signal in self.signals.values():
            # Producer is always a set, to handle pass-through signals
            for producer in signal.producer - set(signal_groups.keys()):
                signal_groups[producer] = []
            for producer in signal.producer:
                signal_groups[producer].append(signal)
        return signal_groups

    def create_selective_groups(self, a_names, b_names):
        """Create groups for the a_list communicating with the b_names

        Returns:
            signal_groups (dict): Signal groups
        """
        signal_groups = {name: {'consumer': [], 'producer': []} for name in a_names}
        for signal in self.signals.values():
            for producer in set(signal.producer):
                if producer in a_names and signal.consumers & b_names:
                    signal_groups[producer]['producer'].append(signal)
            for consumer in signal.consumers:
                if consumer in a_names and set(signal.producer) & a_names:
                    signal_groups[consumer]['consumer'].append(signal)
        return signal_groups

    @staticmethod
    def to_yaml(spec, output):
        """Writes spec to yaml file

        Args:
            spec (dict): data for the yaml
            output (Path): file to write to
        """
        with open(output, 'w', encoding="utf-8") as yaml_file:
            yaml = YAML()
            yaml.dump(spec, yaml_file)



# Copyright 2024 Volvo Car Corporation
# Licensed under Apache 2.0.

# -*- coding: utf-8 -*-
"""Module for handling of NVM definitions and for generation of c&h-files defining the memory layout of the NVM."""
import re
import os
import json
from copy import deepcopy

import powertrain_build.build_defs as bd
from powertrain_build.types import byte_size, get_bitmask
from powertrain_build.a2l import A2l
from powertrain_build.problem_logger import ProblemLogger


class NVMDef(ProblemLogger):
    """A class that holds the NVM definitions for the build.

    Provides methods for generating NVM structs and header files.
    The header files includes defines which reference the
    struct elements as variables.
    """

    class WrongTypeException(Exception):
        """WrongTypeException."""

    _allowed_nvm_memory_areas = [
        "NVM_LIST_32",
        "NVM_LIST_16",
        "NVM_LIST_8",
        "NVM_LIST_32_PER",
        "NVM_LIST_16_PER",
        "NVM_LIST_8_PER",
        "NVM_LIST_CRITICAL1",
        "NVM_LIST_CRITICAL2",
    ]

    _nvm_header_footer = "\n#endif /* VCC_NVM_STRUCT_H */\n"

    def __init__(self, project_config, unit_cfg, nvm_vars):
        """Constructor.

        Args:
            project_config (BuildProjConfig): current project configuration
            unit_cfg (UnitConfigs): unit definitions
            nvm_vars (dict): NVM variables from unit configurations (:doc:`UnitConfigs <unit_configs>`)

        """
        super().__init__()
        self._project_config = project_config
        self._nvm_vars = nvm_vars
        # get the size of areas and c-file names
        self._nvm_defs = project_config.get_nvm_defs()
        self._nvm_signals = {}
        self._nvm_memory_areas = tuple()
        self._mem_area_elem_size = {}
        self._area_section = {}
        self._mem_area_pragmas = tuple()
        self._nvm_header_head = self._get_nvm_header_head(unit_cfg)
        src_file_dst = self._project_config.get_src_code_dst_dir()
        self._file_name = os.path.join(src_file_dst, self._nvm_defs["fileName"])
        self._predecl_start = bd.PREDECL_START
        self._predecl_end = bd.PREDECL_END
        self.struct_member_prefix = "_"

        with open(
            os.path.join(project_config.get_root_dir(), self._nvm_defs["baseNvmStructs"]), "r", encoding="utf-8"
        ) as nvm_json:
            self.nvm_definitions = json.load(nvm_json)
        for memory_area in self.nvm_definitions:
            self._add_memory_area(memory_area)
        self._order_nvm_per_area()
        for memory_area in self.nvm_definitions:
            memory_index = self._get_nvm_areas_index(memory_area["name"])
            self.info("Updating %s with index %s", memory_area["name"], memory_index)
            self.nvm_definitions[memory_index]["signals"] = self._get_signal_list(memory_area["signals"])

    @staticmethod
    def _get_nvm_header_head(unit_cfg):
        _nvm_header_head = (
            "/*\n"
            " *  vcc_nvm_struct.h - struct for NVM signals\n"
            " */\n\n"
            "#ifndef VCC_NVM_STRUCT_H\n"
            "#define VCC_NVM_STRUCT_H\n\n"
        )
        _nvm_header_head += unit_cfg.base_types_headers
        return _nvm_header_head

    def _add_memory_area(self, memory_area):
        """Add non-default areas."""
        if memory_area["name"] not in self._allowed_nvm_memory_areas:
            msg = f"Memory area: {memory_area['name']} not allowed"
            self.critical(msg)
            raise NVMDef.WrongTypeException(msg)

        self._nvm_memory_areas = (*self._nvm_memory_areas, memory_area["name"])
        self._mem_area_elem_size.update({memory_area["name"]: byte_size(memory_area["default_datatype"])})

        if memory_area["persistent"]:
            self._area_section.update({memory_area["name"]: "CVC_NVM_P"})
            self._mem_area_pragmas = (*self._mem_area_pragmas, (bd.CVC_NVM_P_START, bd.CVC_NVM_P_END))
        else:
            self._area_section.update({memory_area["name"]: "CVC_NVM"})
            self._mem_area_pragmas = (*self._mem_area_pragmas, (bd.CVC_NVM_START, bd.CVC_NVM_END))

    def _var_to_area(self, var_info):
        """Find the NVM area for the variable.

        Args:
            var_info (dict): a dict with the variable attributes.

        Returns:
            area (str): a string with the area the variable should be
                        defined in.

        """
        mtch = re.match(r"\D+(\d+)?", var_info["type"])
        if mtch.group(1) is None:
            size = "8"
        else:
            size = mtch.group(1)
        if var_info["class"] == "CVC_DISP_NVM":
            return "NVM_LIST_" + size
        if var_info["class"] == "CVC_DISP_NVM_P":
            return "NVM_LIST_" + size + "_PER"
        self.critical("%s is not a NVM variable!", var_info["name"])
        return None

    def _order_nvm_per_area(self):
        """Order the NVM in the areas, based on the definitions."""
        for var, unit_attrib in self._nvm_vars.items():
            for _, var_attribs in unit_attrib.items():
                area = self._var_to_area(var_attribs)
                self._nvm_signals[var] = var_attribs
                self._nvm_signals[var]["area"] = area
                # only use one units definition
                # TODO:Make a nicer solution #
                break

    def nvm_area_iterator(self, nvm_area_name):
        """Get an iterator for NVM 'area'.

        The iterator yields data_type and and variable name for the
        variables defined in the area.

        Args:
            nvm_area_name (str): the name of the NVM area

        Returns:
            tuple: (data_type, variable name, size)

        """
        for nvm_signal_name, nvm_attributes in self._nvm_signals.items():
            if nvm_attributes["area"] == nvm_area_name:
                yield nvm_attributes.get("type"), nvm_signal_name, nvm_attributes.get("width", 1)

    def _a2l_dict(self):
        """Return a dict defining all parameters for a2l-generation."""
        res = {}
        for var, var_attrib in self._nvm_signals.items():
            variable_name = self._get_struct_and_variable_name(var_attrib["area"])[1]
            res[var] = {
                "var": {"var": var, "type": var_attrib["type"], "cvc_type": "CVC_NVM"},
                "a2l_data": {
                    "bitmask": get_bitmask(var_attrib["type"]),
                    "description": var_attrib["description"],
                    "lsb": var_attrib.get("lsb", 1),
                    "max": var_attrib.get("max"),
                    "min": var_attrib.get("min"),
                    "offset": var_attrib.get("offset"),
                    "unit": var_attrib["unit"],
                    "symbol": variable_name,
                    "symbol_offset": var_attrib.get("struct_off"),
                },
            }
            width = var_attrib.get("width", 1)
            if width:
                if isinstance(width, list):
                    res[var]["array"] = width
                elif int(width) > 1:
                    res[var]["array"] = [width]
        return {"vars": res, "function": "VcNvm"}

    @staticmethod
    def _compare_size(signal, widths):
        """Compare a json signal object with widths array.

        :param signal: signal name
        :param widths: array matrix size
        :return: bool True if size is the same

        """
        return signal["x_size"] == widths[0] and signal["y_size"] == widths[1]

    @staticmethod
    def _clean_values(widths):
        """Get clean width values.

        Sometimes width array i set to -1,
        (This data is read out from the simulink model in the matlab scripts,
        -1 is simulink logic for the values should be inherited).
        :param widths:
        :return: a correct widths array
        """
        if widths in ("-1", 1):
            widths = [1, 1]
        return widths

    def _assert_data_type(self, signal, memory_area):
        """Asserts that each signal inside an NVM block is of one of the allowed data types of the block."""
        memory_area_index = self._get_nvm_areas_index(memory_area)
        if "NotApplicable" in self.nvm_definitions[memory_area_index]["allowed_datatypes"]:
            # Special case for NVM_LIST_CRITICAL1 and 2. Any type is acceptable.
            self.debug("Data type is not checked for %s", self.nvm_definitions[memory_area_index]["name"])
            return
        if signal["type"] not in self.nvm_definitions[memory_area_index]["allowed_datatypes"]:
            msg = f"Signal type:{signal['type']} for {signal['name']} not allowed in area:{memory_area}"
            self.critical(msg)
            raise NVMDef.WrongTypeException(msg)

    def _compare_signal(self, signal, data_type, widths, persistent_signal):
        """Compare signal and warn if there is a mismatch in data type or size.

        :param signal: signal object from nvm_structs.json file
        :param data_type: str
        :param widths: array matrix size
        :return: True if the signal is the same in both json struct and values read from simulink model.

        """
        if data_type != signal["type"]:
            log = self.critical if persistent_signal else self.warning
            log("NVM signal: %s type mismatch %s != %s ", signal["name"], data_type, signal["type"])
            return False
        if not self._compare_size(signal, widths):
            self.critical(
                "signal:%s  size mismatch %s != %s,%s ", signal["name"], widths, signal["x_size"], signal["y_size"]
            )
            return False

        return True

    def _get_nvm_areas_index(self, area):
        """Get array index of nvm area.

        :param area: wanted nvm area as string
        :return: (int) nvm areas index in self.nvm_definitions
        """
        return next(
            (index for index, nvm_definition in enumerate(self.nvm_definitions) if nvm_definition["name"] == area), -1
        )

    def _find_empty_index(self, memory_index, signal):
        """Find empty index in array of signals.

        :param memory_index: array index (NVM area)
        :param signal: signal to add at empty index
        :return: (int) empty index, (-1 if not found)
        """
        for signal_index in range(len(self.nvm_definitions[memory_index]["signals"])):
            signal_candidate = self.nvm_definitions[memory_index]["signals"][signal_index]
            if self._compare_size(signal, (signal_candidate["x_size"], signal_candidate["y_size"])) and (
                signal_candidate["name"].startswith("Pos_") or signal_candidate["name"].startswith("Position_")
            ):
                return signal_index
        return -1

    def _add_signal_to_nvm_struct(self, memory_index, signal_name, data_type, widths):
        """Add a signal to nvm struct at first empty index.

        :param memory_index: array index (NVM area)
        :param signal_name: str signal_nameiable name
        :param data_type: str
        :param widths: array matrix size
        """
        if widths == "-1":
            widths = [1, 1]
        elif widths == "-":
            self.critical("Bad widths value (%s), cannot add: %s", widths, signal_name)
            return

        new_signal = self._get_memory_area_signal_template(memory_index)
        new_signal["name"] = signal_name
        new_signal["type"] = data_type
        new_signal["x_size"] = int(widths[0])
        new_signal["y_size"] = int(widths[1])
        signal_index = self._find_empty_index(memory_index, new_signal)
        if signal_index > -1:
            self.debug("Add at empty signal_index %s", signal_index)
            self._update_nvm_signal(memory_index, signal_index, new_signal)
        else:
            self.nvm_definitions[memory_index]["signals"].append(new_signal)

    def _get_signals_in_nvm_structs(self):
        signals_in_nvm_structs = {}
        for memory_area in self._nvm_memory_areas:
            memory_area_index = self._get_nvm_areas_index(memory_area)
            signals = self.nvm_definitions[memory_area_index]["signals"]
            for signal_index, signal in enumerate(signals):
                signals_in_nvm_structs.update(
                    {signal["name"]: {"memory_area": memory_area, "signal_index": signal_index}}
                )
        return signals_in_nvm_structs

    def _dummify_nvm_signal(self, memory_area, signal_index):
        dummy_name = f"Pos_{memory_area}bit{'_P' if memory_area.endswith('PER') else ''}_{signal_index}"
        memory_index = self._get_nvm_areas_index(memory_area)
        self.nvm_definitions[memory_index]["signals"][signal_index]["name"] = dummy_name
        self.debug("Name of signal in nvm_structs.json was set to %s", dummy_name)

    def _get_nvm_signal(self, memory_index, signal_index):
        return self.nvm_definitions[memory_index]["signals"][signal_index]

    def _update_nvm_signal(self, memory_index, signal_index, signal):
        self.debug("Replacing %s with %s", self.nvm_definitions[memory_index]["signals"][signal_index], signal)
        self.nvm_definitions[memory_index]["signals"][signal_index] = signal

    def _update_duplicate_nvm_definition(self, nvm_structs_signals, model_signal, model_memory_area):
        """Check if already existing nvm signal is correctly configured."""
        nvm_structs_memory_area = nvm_structs_signals[model_signal["name"]]["memory_area"]
        nvm_structs_signal_index = nvm_structs_signals[model_signal["name"]]["signal_index"]
        nvm_structs_memory_index = self._get_nvm_areas_index(nvm_structs_memory_area)
        nvm_structs_signal = self._get_nvm_signal(nvm_structs_memory_index, nvm_structs_signal_index)
        model_memory_index = self._get_nvm_areas_index(model_memory_area)
        model_signal_widths = (model_signal["x_size"], model_signal["y_size"])

        if "NVM_LIST_CRITICAL" in nvm_structs_memory_area:
            # In this case, we need to update the signal memory area.
            # _var_to_area cannot guess the right area for signals located in the critical sections due to varying data
            # types
            self.debug("Keeping signal %s in %s", model_signal["name"], nvm_structs_memory_area)
            self._nvm_signals[model_signal["name"]]["area"] = nvm_structs_memory_area
            return

        persistent_signal = "_PER" in model_memory_area or "_PER" in nvm_structs_memory_area
        if model_memory_area != nvm_structs_memory_area:
            self.warning("NVM signal memory area mismatch for signal: '%s'", model_signal["name"])
            self.warning("nvm_structs.json: '%s' - Memory area: %s", nvm_structs_memory_area, model_memory_area)
            self._dummify_nvm_signal(nvm_structs_memory_area, nvm_structs_signal_index)
            self._add_signal_to_nvm_struct(
                model_memory_index, model_signal["name"], model_signal["type"], model_signal_widths
            )
        elif not self._compare_signal(nvm_structs_signal, model_signal["type"], model_signal_widths, persistent_signal):
            if persistent_signal:
                self.warning("NVM signal type or size mismatch for signal: '%s'", model_signal["name"])
                self._dummify_nvm_signal(nvm_structs_memory_area, nvm_structs_signal_index)
                self._add_signal_to_nvm_struct(
                    model_memory_index, model_signal["name"], model_signal["type"], model_signal_widths
                )
                self.warning("%s relocated to fitting position", model_signal["name"])
            else:
                self._update_nvm_signal(nvm_structs_memory_index, nvm_structs_signal_index, model_signal)

    def _get_memory_area_signal_template(self, memory_area_index):
        memory_area = self.nvm_definitions[memory_area_index]
        signals = memory_area["signals"]
        if signals:
            return signals[0].copy()

        return {"name": "", "type": "", "x_size": "", "y_size": ""}

    def _update_nvm_base_struct(self):
        """Gather and update all nvm signals.

        1) Updates existing signals,
        2) Deletes/Dummy declares non- and persistent unused signals respectively,
        3) Adds new signals,
        in the nvm struct.
        """

        def remove_unused_signals_from_nvm_struct():
            """Remove unused signals from the nvm struct.

            The signals are listed in an array and accessed through calculated indices, therefore,
            all unused signals must be removed at once.
            """
            for mem_area in self._nvm_memory_areas:
                mem_area_index = self._get_nvm_areas_index(mem_area)
                unused_signals_in_area = [
                    signal_name
                    for signal_name, signal_info in unused_nvm_structs_signals.items()
                    if signal_info["memory_area"] == mem_area
                ]

                # Cannot remove signals from the persistent memory areas,
                # however, they can be exchanged for dummy signals
                if "_PER" in self.nvm_definitions[mem_area_index]["name"]:
                    for unused_signal_name in unused_signals_in_area:
                        if unused_signal_name.startswith("Pos"):
                            self.debug("Found unused signal: '%s' in memory area: %s", unused_signal_name, mem_area)
                        else:
                            self.warning("Found unused signal: '%s' in memory area: %s", unused_signal_name, mem_area)
                        self._dummify_nvm_signal(
                            mem_area, unused_nvm_structs_signals[unused_signal_name]["signal_index"]
                        )
                else:
                    self.nvm_definitions[mem_area_index]["signals"] = [
                        signal
                        for signal in self.nvm_definitions[mem_area_index]["signals"]
                        if signal["name"] not in unused_signals_in_area
                    ]

        nvm_structs_signals = self._get_signals_in_nvm_structs()
        unused_nvm_structs_signals = self._get_signals_in_nvm_structs()
        new_nvm_structs_signals = []

        for nvm_memory_area in self._nvm_memory_areas:
            memory_area_index = self._get_nvm_areas_index(nvm_memory_area)
            if "NVM_LIST_CRITICAL" in nvm_memory_area:
                # NVM_LIST_CRITICAL.* only exists in some projects
                continue
            if memory_area_index == -1:
                self.critical("NVM area not found for: %s", nvm_memory_area)
                continue
            for nvm_data_type, nvm_signal_name, nvm_widths in self.nvm_area_iterator(nvm_memory_area):
                nvm_widths = self._clean_values(nvm_widths)
                nvm_signal = self._get_memory_area_signal_template(memory_area_index)
                nvm_signal["name"] = nvm_signal_name
                nvm_signal["type"] = nvm_data_type
                nvm_signal["x_size"] = int(nvm_widths[0])
                nvm_signal["y_size"] = int(nvm_widths[1])
                if nvm_signal_name in nvm_structs_signals:
                    self.debug("Deleting %s", nvm_signal_name)
                    del unused_nvm_structs_signals[nvm_signal_name]
                    self._update_duplicate_nvm_definition(nvm_structs_signals, nvm_signal, nvm_memory_area)
                else:
                    new_nvm_structs_signals.append((memory_area_index, nvm_signal["name"], nvm_data_type, nvm_widths))
        remove_unused_signals_from_nvm_struct()

        for new_signal in new_nvm_structs_signals:
            self._add_signal_to_nvm_struct(*new_signal)

    def _generate_nvm_structs_updated(self):
        self.info("Start generating updated nvm json file")

        nvm_structs_updated_path = os.path.abspath(
            os.path.join(self._project_config.get_root_dir(), "output", "nvm_structs_updated.json")
        )

        self.info("Created %s", nvm_structs_updated_path)

        with open(nvm_structs_updated_path, "w", encoding="utf-8") as nvm_structs_file:
            json.dump(self.nvm_definitions, nvm_structs_file, indent=4)

    def _get_struct_and_variable_name(self, memory_area, skip_prefix=False):
        """Get the name of the memory area struct and variable definitions."""
        prefix = self._project_config.get_scheduler_prefix()
        struct_name = f"{prefix}{memory_area}" if not skip_prefix else memory_area
        if self._project_config.get_code_generation_config("useCamelCaseForNvmVariables"):
            variable_name = "s" + "".join([part[0].upper() + part.lower()[1:] for part in struct_name.split("_")])
        else:
            variable_name = struct_name.lower()
        return struct_name, variable_name

    def _get_signal_and_struct_defines(self):
        """Get NVM signal and struct defines."""
        defines = []
        struct_defines = []
        for memory_area in self._nvm_memory_areas:
            struct_name, variable_name = self._get_struct_and_variable_name(memory_area)
            memory_area_index = self._get_nvm_areas_index(memory_area)
            section = self._area_section[memory_area]
            elem_size = self._mem_area_elem_size[memory_area]
            struct_defines.append(f'#include "{section}_START.h"\n')
            struct_defines.append(f"struct {struct_name} {{\n")

            struct_off = 0
            signals = self.nvm_definitions[memory_area_index]["signals"]
            for signal in signals:
                self._assert_data_type(signal, memory_area)
                struct_defines.append(f'   {signal["type"]:7} {self.struct_member_prefix}{signal["name"]}')
                size_string = ""
                size = max(signal["x_size"], 1) * max(signal["y_size"], 1)
                if size > 1:
                    if signal["x_size"] > 1:
                        size_string += f'[{signal["x_size"]}]'
                    if signal["y_size"] > 1:
                        if signal["x_size"] < 2:
                            self.critical("NVM signal size incorrect. x_size should not be 1 if y_size > 1")
                            size_string += "[1]"
                        size_string += f'[{signal["y_size"]}]'
                    struct_defines.append(size_string)
                struct_defines.append(";\n")

                if signal["name"] in self._nvm_signals:
                    self._nvm_signals[signal["name"]]["struct_off"] = struct_off

                struct_off += size * byte_size(signal["type"])

                defines.append(
                    "#define "
                    f"{signal['name']:40} "
                    f"{variable_name}.{self.struct_member_prefix}{signal['name']}\n"
                )

            max_nr_signals = self.nvm_definitions[memory_area_index]["size"]
            tot_memory = max_nr_signals * elem_size
            free = tot_memory - struct_off
            if free > 0:
                struct_defines.append(
                    f'   {self.nvm_definitions[memory_area_index]["default_datatype"]} '
                    f'unused[{int(free / elem_size)}];\n'
                )
            elif free < 0:
                self.critical("NVM area %s overrun!", self.nvm_definitions[memory_area_index]["name"])
            struct_defines.append(f"}}; /* {struct_off} bytes used of {tot_memory} */\n")
            struct_defines.append(f'#include "{section}_END.h"\n\n')

        return defines, struct_defines

    def _generate_nvm_config_headers(self):
        """Generate nvm config h file."""
        self.info("Start generating nvm header file")
        defines, struct_defines = self._get_signal_and_struct_defines()

        externals = [f'#include "{self._predecl_start}"\n']
        for area in self._nvm_memory_areas:
            struct_name, variable_name = self._get_struct_and_variable_name(area)
            externals.append(f"extern struct {struct_name} {variable_name};\n")
        externals.append(f'#include "{self._predecl_end}"\n\n')

        with open(self._file_name + ".h", "w", encoding="utf-8") as hptr:
            hptr.write(self._nvm_header_head)
            hptr.writelines(struct_defines)
            hptr.writelines(externals)
            hptr.write("\n")
            hptr.writelines(defines)
            hptr.write(self._nvm_header_footer)

    def _generate_nvm_config_source(self):
        """Generate the c-file containing the NVM definition."""
        # TODO: Add memory from previous builds!!! and mark old positions #
        self.info("Start generating nvm source file")
        with open(self._file_name + ".c", "w", encoding="utf-8") as cptr:
            cptr.write(f'#include "{self._nvm_defs["fileName"]}.h"\n\n')
            for area, pragma in zip(self._nvm_memory_areas, self._mem_area_pragmas):
                struct_name, variable_name = self._get_struct_and_variable_name(area)
                cptr.write(f'#include "{pragma[0]}"\n')
                cptr.write(f"struct {struct_name} {variable_name};\n")
                cptr.write(f'#include "{pragma[1]}"\n\n')

    def _generate_nvm_config_a2l(self):
        """Generate the a2l-file describing the NVM definition."""
        self.info("Start generating nvm a2l file")
        a2l_dict = self._a2l_dict()
        a2l = A2l(a2l_dict, self._project_config)
        a2l.gen_a2l(self._file_name + ".a2l")

    def generate_nvm_config_files(self, no_nvm_a2l):
        """Generate all files for variables in the NVM definition.

        Args:
            no_nvm_a2l (bool): Do not generate a2l file.
        """
        self._update_nvm_base_struct()
        self._generate_nvm_structs_updated()
        self._generate_nvm_config_headers()
        self._generate_nvm_config_source()
        if not no_nvm_a2l:
            self._generate_nvm_config_a2l()

    @staticmethod
    def _get_signal_list(signals):
        return signals if isinstance(signals, list) else [signals]


class ZCNVMDef(NVMDef):
    """A class for handling of ZC NVM definitions."""

    def __init__(self, project_config, unit_cfg, nvm_vars):
        """Init.

        Args:
            project_config (BuildProjConfig): current project configuration.
            unit_cfg (UnitConfigs): Unit configurations.
            nvm_vars (dict): NVM variables from unit configurations.
        """
        super().__init__(project_config, unit_cfg, nvm_vars)
        self.struct_member_prefix = "e_"
        self._valid_nvm_definitions = None
        self._update_nvm_base_struct()
        prefix = self._project_config.get_scheduler_prefix()
        self.project_nvm_definitions = {
            f"{prefix}{item['name']}": item for item in self.nvm_definitions
        }
        self._predecl_start = bd.PREDECL_NVM_START
        self._predecl_end = bd.PREDECL_NVM_END
        self.init_function = f"void {prefix}VcNvmInit(void)"
        self.main_function = f"void {prefix}VcNvm(void)"

    @property
    def valid_nvm_definitions(self):
        """Get the valid NVM definitions."""
        return self._valid_nvm_definitions

    @valid_nvm_definitions.setter
    def valid_nvm_definitions(self, yaml_nvm_definitions):
        """Return a set of NVM definitions appearing in both the project and the project yaml file.

        Args:
            yaml_nvm_definitions (dict): NVM lists listed in the NVM configuration yaml file.
        Returns:
            valid_nvm_definitions (dict): Validated NVM definitions,
                listed in both NVM configuration yaml file as well as project.
        """
        local_yaml_nvm_definitions = deepcopy(yaml_nvm_definitions)
        nvms_not_in_project = set(local_yaml_nvm_definitions.keys()) - set(self.project_nvm_definitions.keys())
        nvms_not_in_yaml = set(self.project_nvm_definitions.keys()) - set(local_yaml_nvm_definitions.keys())
        for key in nvms_not_in_project:
            self.warning(f'Ignoring NVM definition {key} since it does not appear in nvm_structs.json.')
            del local_yaml_nvm_definitions[key]
        for key in nvms_not_in_yaml:
            self.warning(f'Ignoring NVM definition {key} since it does not appear in the project NVM yaml file.')
        self._valid_nvm_definitions = local_yaml_nvm_definitions

    def _update_header_and_footer(self):
        name = self._project_config.get_composition_config("softwareComponentName")
        use_rte_nvm_structs = self._project_config.get_code_generation_config("useRteNvmStructs")
        self._nvm_header_head += f'#include "Rte_{name}.h"\n'
        if use_rte_nvm_structs:
            self._nvm_header_head += '#include "Rte_Type.h"\n\n'
        self._nvm_header_footer = (
            "\n"
            f'#include "{bd.PREDECL_CODE_ASIL_D_START}"\n'
            f"{self.init_function};\n"
            f"{self.main_function};\n"
            f'#include "{bd.PREDECL_CODE_ASIL_D_END}"\n'
            f"{self._nvm_header_footer}"
        )

    def _generate_rte_type_nvm_config_headers(self):
        """Generate NVM config header file using RTE struct definitions."""
        self.info("Start generating nvm header file")
        defines = self._get_signal_and_struct_defines()[0]
        externals = [f'#include "{self._predecl_start}"\n']
        for area in self.valid_nvm_definitions.keys():
            struct_name, variable_name = self._get_struct_and_variable_name(area, skip_prefix=True)
            externals.append(f"extern dt_{struct_name} {variable_name};\n")
        externals.append(f'#include "{self._predecl_end}"\n\n')
        with open(self._file_name + ".h", "w", encoding="utf-8") as hptr:
            hptr.write(self._nvm_header_head)
            hptr.writelines(externals)
            hptr.writelines(defines)
            hptr.write(self._nvm_header_footer)

    def _generate_rte_type_nvm_config_source(self):
        """Generate the c-file containing the NVM definition using RTE struct definitions."""
        # TODO: Add memory from previous builds!!! and mark old positions #
        self.info("Start generating nvm source file")
        with open(self._file_name + ".c", "w", encoding="utf-8") as cptr:
            cptr.write(f'#include "{self._nvm_defs["fileName"]}.h"\n\n')
            for area, pragma in zip(self._nvm_memory_areas, self._mem_area_pragmas):
                struct_name, variable_name = self._get_struct_and_variable_name(area)
                if struct_name in self.valid_nvm_definitions:
                    cptr.write(f'#include "{pragma[0]}"\n')
                    cptr.write(f"dt_{struct_name} {variable_name};\n")
                    cptr.write(f'#include "{pragma[1]}"\n\n')

    def _append_nvm_rte_function_calls(self):
        """Append the NVM RTE function calls to the NVM config source file."""
        init_function_lines = [
            f'#include "{bd.CVC_CODE_ASIL_D_START}"\n'
            f"{self.init_function} {{\n",
            "  //Call this function in the SWC init runnable\n"
        ]
        main_function = [
            f'#include "{bd.CVC_CODE_ASIL_D_START}"\n'
            f"{self.main_function} {{\n",
            "  //Call this function in the SWC main runnable\n",
        ]
        footer = [
            "}\n",
            f'#include "{bd.CVC_CODE_ASIL_D_END}"\n',
        ]

        main_function_lines = []

        for nvm_name, nvm_data in self.valid_nvm_definitions.items():
            if nvm_data["type"] not in ["type1", "type2"]:
                self.critical(f"Unknown NVM type: {nvm_data['type']} for NVM: {nvm_name}")
                continue
            if nvm_data["method"] not in ["DIRECT-CALL", "NVSWCOMPONENT"]:
                self.critical(f"Unknown NVM method: {nvm_data['method']} for NVM: {nvm_name}")
                continue

            if nvm_data["type"] == "type2":
                self.warning(
                    f'{nvm_name} is a type2 {nvm_data["method"]}. Due to inherent risk type2 write calls are '
                    'not generated automatically, they have to be written manually and added to source code.'
                )
                main_function_lines.append(
                    f"  // Nvm {nvm_name} is type 2, add call in manually written source code.\n"
                )
            elif nvm_data["type"] == "type1":
                variable_name = self._get_struct_and_variable_name(nvm_name, skip_prefix=True)[1]
                if nvm_data["method"] == "DIRECT-CALL":
                    pim_call = f"Rte_Pim_{nvm_name}()"
                    main_function_lines.append(f"  *{pim_call} = {variable_name};\n")
                    init_function_lines.append(f"  {variable_name} = *{pim_call};\n")
                elif nvm_data["method"] == "NVSWCOMPONENT":
                    init_function_lines.append(
                        f"  Rte_Read_{nvm_name.upper()}_{nvm_name.upper()}(&{variable_name});\n"
                    )
                    main_function.append(
                        f"  Rte_Write_{nvm_name.upper()}_{nvm_name.upper()}(&{variable_name});\n"
                    )

        output = init_function_lines + footer + ["\n"] + main_function + main_function_lines + footer
        with open(self._file_name + ".c", mode="a", encoding="utf-8") as cptr:
            cptr.writelines(output)

    def generate_nvm_rte_files(self):
        """Generate the NVM RTE files."""
        if self.valid_nvm_definitions is None:
            self.critical('Valid NVM definitions not set. Cannot generate NVM RTE files.')
            return

        use_rte_nvm_structs = self._project_config.get_code_generation_config("useRteNvmStructs")

        self._update_header_and_footer()
        self._generate_nvm_structs_updated()
        if use_rte_nvm_structs:
            self._generate_rte_type_nvm_config_headers()
            self._generate_rte_type_nvm_config_source()
        else:
            self._generate_nvm_config_headers()
            self._generate_nvm_config_source()
        self._append_nvm_rte_function_calls()



# Copyright 2024 Volvo Car Corporation
# Licensed under Apache 2.0.

# -*- coding: utf-8 -*-

"""Python module used for calculating interfaces for CSP"""


import argparse
import os
import re
import sys
from itertools import product
from pathlib import Path
from textwrap import dedent
from typing import List, Optional

import git

from powertrain_build.interface.application import Application, Model, get_active_signals
from powertrain_build.interface.ems import CsvEMS
from powertrain_build.lib import logger

LOGGER = logger.create_logger("Check interface")


def process_app(config):
    """Get an app specification for the current project

    Entrypoint for external scripts.

    Args:
        config (pathlib.Path): Path to the ProjectCfg.json
    Returns:
        app (Application): pybuild project
    """
    app = Application()
    app.parse_definition(config)
    return app


def model_app_consistency(model, app_models, app, errors):
    """Compare model signal interface with list of models.

    Args:
        model (Model): model to compare against application
        app_models (list(Model)): list of models to compare with
        app (Application): pybuild project
        errors (dict): Object for counting errors of different types
    """

    for compare_model in app_models:
        LOGGER.debug("Comparing %s with %s in %s", model.name, compare_model.name, app.name)
        active_model_outsignals = get_active_signals(model.outsignals, app.pybuild["feature_cfg"])
        active_model_insignals = get_active_signals(model.insignals, app.pybuild["feature_cfg"])
        active_compare_outsignals = get_active_signals(compare_model.outsignals, app.pybuild["feature_cfg"])
        active_compare_insignals = get_active_signals(compare_model.insignals, app.pybuild["feature_cfg"])
        check_signals(
            active_model_insignals,
            active_compare_outsignals,
            errors,
            [app.name, model.name],
            [app.name, compare_model.name],
        )
        check_signals(
            active_model_outsignals,
            active_compare_insignals,
            errors,
            [app.name, model.name],
            [app.name, compare_model.name],
        )


def check_internal_signals(app, model_names=None):
    """Look for all internal signal mismatches.

    Args:
        app (Application): pybuild project
        model_names (list(Model)): models based on parsed config jsons
    Returns:
        serious_mismatch (bool): A serious mismatch was found
    """
    serious_mismatch = False
    LOGGER.debug("Checking internal signals")
    LOGGER.debug("Checking against %s", app.signals)
    errors = {"type": 0, "range": 0, "unit": 0, "width": 0}
    app_models = app.get_models()
    for signal in app.signals:
        LOGGER.debug(signal.properties)

    for model in app_models:
        if model_names is not None and model.name not in model_names:
            LOGGER.debug("Skipping %s", model.name)
            continue
        LOGGER.debug("Checking %s in %s", model.name, app.name)
        active_insignals = get_active_signals(model.insignals, app.pybuild["feature_cfg"])
        insignal_mismatch = check_signals(active_insignals, app.signals, errors, [app.name, model.name], [app.name])
        active_outsignals = get_active_signals(model.outsignals, app.pybuild["feature_cfg"])
        outsignal_mismatch = check_signals(active_outsignals, app.signals, errors, [app.name, model.name], [app.name])
        if insignal_mismatch or outsignal_mismatch:
            serious_mismatch = True
            model_app_consistency(model, app_models, app, errors)
        # Only compare with all models if a mismatch is found
    LOGGER.debug("Total errors: %s", errors)
    return serious_mismatch


def check_models_generic(all_models, model_names, emses):
    """Check filtered models against all models and external interfaces."""
    serious_mismatch = False
    for model in all_models:
        LOGGER.info("Checking signals attributes for %s", model.name)
        if model.name not in model_names:
            continue
        errors = {"type": 0, "range": 0, "unit": 0, "width": 0}
        LOGGER.debug("Checking internal signals for %s", model.name)
        for corresponding_model in all_models:
            serious_mismatch |= check_signals(
                model.insignals, corresponding_model.outsignals, errors, [model.name], [corresponding_model.name]
            )
            serious_mismatch |= check_signals(
                model.outsignals, corresponding_model.insignals, errors, [model.name], [corresponding_model.name]
            )
        if emses:
            LOGGER.debug("Checking external signals for %s", model.name)
        for ems in emses:
            serious_mismatch |= check_signals(
                model.insignals, ems.outsignals, errors, [model.name], [ems.name]
            )
            serious_mismatch |= check_signals(
                model.outsignals, ems.insignals, errors, [model.name], [ems.name]
            )
        LOGGER.debug("Total errors for %s: %s", model.name, errors)
    return serious_mismatch


def get_all_models(model_root):
    """Find, filter and parse all model configurations."""
    LOGGER.info("Parsing all models")
    prefix = "config_"
    suffix = ".json"
    models = []
    for dirpath, _, filenames in os.walk(model_root):
        dirpath = Path(dirpath)
        for filename in [f for f in filenames if f.startswith(prefix) and f.endswith(suffix)]:
            name = filename[len(prefix): -len(suffix)]
            if name == dirpath.parent.stem:
                model = Model(None)
                model.parse_definition((name, Path(dirpath, filename)))
                models.append(model)
    return models


def get_projects(root, project_names):
    """Find, parse and filter all project configurations."""
    LOGGER.info("Parsing all projects")
    projects = []
    for dirpath, _, filenames in os.walk(root):
        dirpath = Path(dirpath)
        for filename in [f for f in filenames if f == "ProjectCfg.json"]:
            config = Path(dirpath, filename)
            app = Application()
            app_name = app.get_name(config)
            if project_names is not None and app_name not in project_names:
                if config.parent.stem not in project_names:
                    LOGGER.info("%s or %s does not match %s", app_name, config.parent.stem, project_names)
                    continue
            app.parse_definition(config)
            if app.pybuild["build_cfg"].has_yaml_interface:
                LOGGER.warning("Interface checks for yaml-interface projects are not implemtented yet")
                LOGGER.info("Adding empty interface for %s", app_name)
                projects.append((app, None))
            else:
                ems = CsvEMS()
                ems.parse_definition(config)
                projects.append((app, ems))
    return projects


def correct_type(left_spec, right_spec):
    """Check if the type is the same in two specifications.

    Args:
        left_spec (dict): Signal specification
        right_spec (dict): Signal specification to compare with
    Returns:
        matches (bool): Spec1 and Spec2 has the same type
    """
    return left_spec["type"] == right_spec["type"]


def correct_attribute(left_spec, right_spec, attribute, default=None, check_bool=True):
    """Check attributes other than type.

    Args:
        left_spec (dict): Signal specification
        right_spec (dict): Signal specification to compare with
        attribute (string): Attribute to check
        default (value): Default value for the attribute (default: None)
        check_bool (bool): Check signals of type Bool (default: True)
    Returns:
        matches (bool): Spec1 and Spec2 has the same value for the attribute
    """

    def _format(value):
        if isinstance(value, str):
            value = value.strip()
            if re.fullmatch("[+-]?[0-9]+", value):
                value = int(value)
            elif re.fullmatch("[+-]?[0-9]+[0-9.,eE+]*", value):
                value = float(value.replace(",", "."))
        return value

    if not check_bool and left_spec["type"] == "Bool":
        return True
    return _format(left_spec.get(attribute, default)) == _format(right_spec.get(attribute, default))


def found_mismatch(name, left_spec, right_spec, attribute, left_path, right_path):
    """Handle finding a mismatch.

    Args:
        name (string): Name of signal
        left_spec (dict): Spec of signal
        right_spec (dict): Signal specification to compare with
        attribute (string): Attribute to check
        left_path (list(str)): Path for where the left signals' definitions come from
        right_path (list(str)): Path for where the right signals' definitions come from
    """
    if attribute in ["type", "width"]:
        # TODO: Add more properties as serious when the interfaces are more cleaned up
        LOGGER.error(
            "%s has %ss: %s in %s and %s in %s",
            name,
            attribute,
            left_spec.get(attribute),
            left_path,
            right_spec.get(attribute),
            right_path,
        )
        return True
    LOGGER.info(
        "%s has %ss: %s in %s and %s in %s",
        name,
        attribute,
        left_spec.get(attribute),
        left_path,
        right_spec.get(attribute),
        right_path,
    )
    return False


def check_external_signals(ems, app, model_names=None):
    """Look for external signal mismatches.

    Args:
        ems (CsvEMS): Parsed signal interface cvs:s
        app (Application): Parsed project config
        model_names (list(Model)): models based on parsed config jsons
    Returns:
        serious_mismatch (bool): A serious mismatch was found
    """
    serious_mismatch = False

    LOGGER.debug("Checking insignals")
    errors = {"type": 0, "range": 0, "unit": 0, "width": 0}
    app_models = app.get_models()
    for model in app_models:
        if model_names is not None and model.name not in model_names:
            LOGGER.debug("Skipping %s in %s", model.name, app.name)
            continue
        LOGGER.debug("Checking %s in %s", model.name, app.name)
        serious_mismatch |= check_signals(
            get_active_signals(model.insignals, app.pybuild["feature_cfg"]),
            ems.outsignals,
            errors,
            [app.name, model.name],
            [ems.name],
        )
        serious_mismatch |= check_signals(
            get_active_signals(model.outsignals, app.pybuild["feature_cfg"]),
            ems.insignals,
            errors,
            [app.name, model.name],
            [ems.name],
        )
    LOGGER.debug("Total errors: %s", errors)
    return serious_mismatch


def check_signals(left_signals, right_signals, errors, left_path=None, right_path=None):
    """Compares insignals from one system with the outsignals of another.

    Args:
        left_signals (list(Signal)): Insignals of one system such as a model
        right_signals (list(Signal)): Outsignals of system to compare with
        errors (dict): Object for counting errors of different types
        left_path (list(str)): Path for where the left signals' definitions come from
        right_path (list(str)): Path for where the right signals' definitions come from
    Returns:
        serious_mismatch (bool): A serious mismatch was found
    """
    left_path = [] if left_path is None else left_path
    right_path = [] if right_path is None else right_path
    serious_mismatch = False
    LOGGER.debug("Checking from %s", left_signals)
    LOGGER.debug("Checking against %s", right_signals)
    for (left_signal, right_signal) in [
        (left, right) for left, right in product(left_signals, right_signals) if left.name == right.name
    ]:
        LOGGER.debug("Comparing %s and %s", left_signal, right_signal)
        left_properties = left_signal.properties
        right_properties = right_signal.properties
        LOGGER.debug("Properties left: %s", left_properties)
        LOGGER.debug("Properties right: %s", right_properties)
        if not correct_type(left_properties, right_properties):
            serious_mismatch |= found_mismatch(
                left_signal.name, left_properties, right_properties, "type", left_path, right_path
            )
            errors["type"] += 1
        if not correct_attribute(left_properties, right_properties, "min", check_bool=False):
            serious_mismatch |= found_mismatch(
                left_signal.name, left_properties, right_properties, "min", left_path, right_path
            )
            errors["range"] += 1
        if not correct_attribute(left_properties, right_properties, "max", check_bool=False):
            serious_mismatch |= found_mismatch(
                left_signal.name, left_properties, right_properties, "max", left_path, right_path
            )
            errors["range"] += 1
        if not correct_attribute(left_properties, right_properties, "unit", default="", check_bool=False):
            serious_mismatch |= found_mismatch(
                left_signal.name, left_properties, right_properties, "unit", left_path, right_path
            )
            errors["unit"] += 1
        if not correct_attribute(left_properties, right_properties, "width", default=1):
            serious_mismatch |= found_mismatch(
                left_signal.name, left_properties, right_properties, "width", left_path, right_path
            )
            errors["width"] += 1
    return serious_mismatch


PARSER_HELP = dedent(r"""
    Checks attributes and existence of signals

    Produced but not consumed signals are giving warnings
    Consumed but not produced signals are giving errors

    Attributes checked are: types, ranges, units and widths
    Mismatches in types or widths give errors
    Mismatches in min, max or unit gives warnings

    Examples:
    py -3.6 -m powertrain_build.check_interface models_in_projects <Projects> <Models/ModelGroup>\
            --projects <ProjectOne> <ProjectTwo>
    Checks models in Models/ModelGroup against ProjectOne and ProjectTwo in the folder Projects

    py -3.6 -m powertrain_build.check_interface models <Models> --models <ModelOne> <ModelTwo>
    Checks models ModelOne and ModelTwo against all other models in the folder Models

    py -3.6 -m powertrain_build.check_interface projects <Projects> \
            --projects ProjectOne ProjectTwo ProjectThree
    Checks the interfaces of ProjectOne, ProjectTwo and ProjectThree in the folder Projects
""").strip()


def configure_parser(parser: argparse.ArgumentParser):
    """Configure arguments in parser."""
    subparsers = parser.add_subparsers(
        help="help for subcommand",
        dest="mode",
        required=True,
    )

    # create the parser for the different commands
    model = subparsers.add_parser(
        "models",
        description=dedent("""
            Check models independently of projects.

            All signals are assumed to be active.
            Any signal that gives and error is used in a model but is not produced in any model or project
            interface.
        """).strip(),
    )
    add_model_args(model)
    model.set_defaults(func=model_check)

    project = subparsers.add_parser(
        "projects",
        description=dedent("""
            Check projects as a whole.

            It checks all models intenally and the SPM vs the interface.
        """).strip(),
    )
    add_project_args(project)
    project.set_defaults(func=projects_check)

    models_in_projects = subparsers.add_parser(
        "models_in_projects",
        description=dedent("""
            Check models specifically for projects.

            Codeswitches are used to determine if the signals are produced and consumed in each model.
        """).strip(),
    )
    add_project_args(models_in_projects)
    add_model_args(models_in_projects)
    models_in_projects.add_argument("--properties", help="Check properties such as type", action="store_true")
    models_in_projects.add_argument("--existence", help="Check signal existence consistency", action="store_true")
    models_in_projects.set_defaults(func=models_in_projects_check)


def add_project_args(parser: argparse.ArgumentParser):
    """Add project arguments to subparser"""
    parser.add_argument("project_root", help="Path to start looking for projects", type=Path)
    parser.add_argument(
        "--projects", help="Name of projects to check. Matches both path and interface name.", nargs="+"
    )


def add_model_args(parser: argparse.ArgumentParser):
    """Add model arguments to subparser"""
    parser.add_argument("model_root", help="Path to start looking for models", type=Path)
    parser.add_argument("--models", help="Name of models to check", nargs="+")
    parser.add_argument("--gerrit", action="store_true", help="Deprecated")
    parser.add_argument("--git", action="store_true", help="Get models to check from git HEAD")


def get_changed_models():
    """Get changed models in current commit."""
    repo = git.Repo()
    changed_files_tmp = repo.git.diff("--diff-filter=d", "--name-only", "HEAD~1")
    changed_files = changed_files_tmp.splitlines()
    changed_models = [m for m in changed_files if m.endswith(".mdl") or m.endswith(".slx")]
    return changed_models


def model_path_to_name(model_paths):
    """Extract model names from a list of model paths."""
    model_names = []
    for model_path in model_paths:
        model_name_with_extension = model_path.split("/")[-1]
        model_name = model_name_with_extension.split(".")[0]
        model_names.append(model_name)
    return model_names


def model_check(args: argparse.Namespace):
    """Entry point for models command."""
    serious_mismatch = False
    all_models = get_all_models(args.model_root)
    if args.models is not None:
        model_names = args.models
    elif args.git or args.gerrit:
        # Still checking args.gerrit due to common-linux-signal_consistency in pt-zuul-jobs
        model_paths = get_changed_models()
        model_names = model_path_to_name(model_paths)
    else:
        model_names = [model.name for model in all_models]

    serious_mismatch |= check_models_generic(all_models, model_names, [])

    if serious_mismatch:
        LOGGER.error("Serious interface errors found.")

    return serious_mismatch


def projects_check(args: argparse.Namespace):
    """Entry point for projects command."""
    serious_mismatch = False
    projects = get_projects(args.project_root, args.projects)
    for app, ems in projects:
        LOGGER.info("Checking interfaces for %s", app.name)
        serious_mismatch |= check_internal_signals(app, None)
        if ems is not None:
            serious_mismatch |= check_external_signals(ems, app, None)

    if serious_mismatch:
        LOGGER.error("Serious interface errors found.")

    return serious_mismatch


def models_in_projects_check(args: argparse.Namespace):
    """Entry point for models_in_projects command."""
    serious_mismatch = False
    projects = get_projects(args.project_root, args.projects)
    LOGGER.debug("Checking projects: %s", projects)
    if args.properties:
        for app, ems in projects:
            serious_mismatch |= check_internal_signals(app, None or args.models)
            if ems is not None:
                serious_mismatch |= check_external_signals(ems, app, None or args.models)
    if args.existence:
        all_models = get_all_models(args.model_root)
        model_names = [model.name for model in all_models] if args.models is None else args.models
        serious_mismatch |= signal_existence(projects, model_names)

    if serious_mismatch:
        LOGGER.error("Serious interface errors found.")

    return serious_mismatch


def signal_existence(projects, model_names):
    """Check which signals are consumed and produced in each project."""
    serious_mismatch = False
    for app, ems in projects:
        app_models = app.get_models()
        LOGGER.info("Checking %s", app.name)
        for project_model in app_models:
            if project_model.name not in model_names:
                continue
            LOGGER.debug("Checking signal existence for %s", project_model.name)
            active_insignals = get_active_signals(project_model.insignals, app.pybuild["feature_cfg"])
            active_outsignals = get_active_signals(project_model.outsignals, app.pybuild["feature_cfg"])
            insignal_matches = {}
            outsignal_matches = {}
            for check_model in app_models:
                signal_match(
                    active_insignals,
                    get_active_signals(check_model.outsignals, app.pybuild["feature_cfg"]),
                    insignal_matches,
                )
                signal_match(
                    active_outsignals,
                    get_active_signals(check_model.insignals, app.pybuild["feature_cfg"]),
                    outsignal_matches,
                )
            if ems is not None:
                signal_match(active_insignals, ems.outsignals, insignal_matches)
                signal_match(active_outsignals, ems.insignals, outsignal_matches)
            for missing_signal in [signal for signal, matched in insignal_matches.items() if not matched]:
                # serious_mismatch = True  # TODO: Activate this code when we want to gate on it.
                LOGGER.warning(
                    "%s is consumed in %s but never produced in %s", missing_signal, project_model.name, app.name
                )
            for missing_signal in [signal for signal, matched in insignal_matches.items() if not matched]:
                LOGGER.debug("%s is consumed in %s and produced in %s", missing_signal, project_model.name, app.name)
            for missing_signal in [signal for signal, matched in outsignal_matches.items() if not matched]:
                LOGGER.info(
                    "%s is produced in %s but never consumed in %s", missing_signal, project_model.name, app.name
                )
            for missing_signal in [signal for signal, matched in outsignal_matches.items() if not matched]:
                LOGGER.debug("%s is consumed in %s and produced in %s", missing_signal, project_model.name, app.name)
    return serious_mismatch


def signal_match(signals_to_check, signals_to_check_against, matches):
    """Check for matches in signal names."""
    for a_signal in signals_to_check:
        matches[a_signal.name] = matches.get(a_signal.name, False)
        for b_signal in signals_to_check_against:
            if b_signal.name == a_signal.name:
                matches[a_signal.name] = True


def main(argv: Optional[List[str]] = None):
    """Main function for stand alone execution."""
    parser = argparse.ArgumentParser(
        description=PARSER_HELP,
        formatter_class=argparse.RawTextHelpFormatter,
    )
    configure_parser(parser)
    args = parser.parse_args(argv)
    args.func(args)


if __name__ == "__main__":
    main(sys.argv[1:])



# Copyright 2024 Volvo Car Corporation
# Licensed under Apache 2.0.

"""Main package of the powertrain-build application."""
from os import environ
from pbr.version import VersionInfo
from pathlib import Path
from powertrain_build.lib import logger, helper_functions
from powertrain_build.environmentcheck import check_python_string

LOGGER = logger.create_logger(__file__)
__version__ = VersionInfo('powertrain-build').release_string()
LOGGER.info('Current powertrain-build version is %s', __version__)
__config_version__ = '0.2.1'
__required_python_lower__ = '3.6'
__required_python_upper__ = '3.13'


workspace = helper_functions.get_repo_root()
requirement_path = Path(
    workspace, 'Script', 'PyTools', 'requirements.txt'
    )
if requirement_path.exists():
    with requirement_path.open("r") as requirement_file:
        expected_package = "powertrain-build==" + __version__
        for line in requirement_file:
            if expected_package in line:
                LOGGER.info('powertrain-build version matched requirements!')
                break
            elif "powertrain-build==" in line and expected_package not in line:
                LOGGER.warning('powertrain-build version does not match requirements!')
                break
else:
    LOGGER.warning('Current repository does not have a requirement file' +
                   ' in expected location: %s', str(requirement_path))

check_python_string(__required_python_lower__, __required_python_upper__)

if environ.get('MatInstl2017') is not None:
    LOGGER.warning('Using environment variable MatInstl2017 is depricated, use MatInstl instead.')



# Copyright 2024 Volvo Car Corporation
# Licensed under Apache 2.0.

"""Module to generate AllSystemInfo.mat for compatibility purposes with DocGen."""

import os
from copy import deepcopy
from numpy import array
from scipy.io import savemat
from powertrain_build.signal_interfaces import SignalInterfaces
from powertrain_build.unit_configs import UnitConfigs
from powertrain_build.lib.helper_functions import merge_dicts
from powertrain_build.problem_logger import ProblemLogger


def _get_signals_by_type(signal_conf, signal_type):
    """Get signals by type ('missing', 'unused', 'inconsistent_defs' or 'multiple_defs').

    Args:
        signal_conf (dict): Configuration from SignalInterfaces with the following format
        {
            "missing": {"signal_name" : ["VED4_GENIII", "VEP4_GENIII"]}
            "unused": {}
            "multiple_defs": {}
            "inconsistent_defs": {}
        }

    Returns:
        dict: with the following format
        {
            "signal_name" : {'VarStatus' : 'Not Used', 'SignalType' : "signal_type"}
        }

    """
    result = {}
    for signal_name in signal_conf[signal_type].keys():
        signal = {signal_name: {'VarStatus': 'Not Used', 'SignalType': signal_type}}
        result.update(signal)
    return result


class GenAllSystemInfo(ProblemLogger):
    """Class to generate AllSystemInfo.mat for compatibility purposes with DocGen."""

    _signal_types = ['missing', 'unused']

    def __init__(self, signal_if, unit_cfg):
        """Class to generate a matlab struct AllSystemInfo for compatibility with the old document generation.

        Args:
            signal_if: an initialized SignalInterfaces object to access signal configurations
        """
        if not isinstance(signal_if, SignalInterfaces):
            raise TypeError
        if not isinstance(unit_cfg, UnitConfigs):
            raise TypeError
        self.signal_if = signal_if
        self.unit_cfg = unit_cfg
        self._signals_without_source = {}
        self._core_ids = {}
        self._file_info = {}
        self._dids = {}

    def __repr__(self):
        """Get string representation of object."""
        return f'if: {self.signal_if}\n uc: {self.unit_cfg}'

    def build(self):
        """Build AllSystemInfo.mat for docgen compatibility."""
        self.info('******************************************************')
        self.info('%s - Getting signals without source', __name__)
        signals_tmp = self.get_signals_without_source()
        for signal_name, signal_data in signals_tmp.items():
            if '.' in signal_name:
                struct_name = signal_name.split('.')[0]
                if struct_name not in self._signals_without_source:
                    self.info(
                        f'{__name__} - Found struct signal: {signal_name}, adding struct name only ({struct_name}). '
                        'Remaining members will be ignored.'
                    )
                    self._signals_without_source[struct_name] = signal_data
            else:
                self._signals_without_source[signal_name] = signal_data

        self.info('%s - Getting CoreIDs', __name__)
        self._core_ids = self.get_core_ids()

        self.info('%s - Generating FileInfo', __name__)
        self._file_info = self._gen_dummy_file_info()

        self.info('%s - Getting DIDs', __name__)
        self._dids = self.get_dids()

        output_path = os.path.abspath(os.path.join('Models', 'CodeGenTmp'))
        self.info('%s - Creating CodeGenTmp directory', __name__)
        self.create_codegen_tmp_dir(output_path)

        self.info('%s - Building AllSystemInfo', __name__)
        self._build_all_system_info(output_path)

    def create_codegen_tmp_dir(self, path):
        """Create CodeGenTmp directory if it does not exist.

        Args:
            path (str): directory to create.
        """
        if not os.path.isdir(path):
            os.mkdir(path)
        else:
            self.info('%s - Directory already exist at %s', __name__, path)

    def _build_all_system_info(self, path):
        """Build AllSystemInfo.mat for DocGen compatibility.

        Args:
            path (str): directory path to place the output file.
        """
        absolute_path = os.path.abspath(os.path.join(path, 'AllSystemInfo.mat'))
        all_system_info = {
            'SystemData': self._signals_without_source,
            'CoreIDs': self._core_ids,
            'FileInfo': self._file_info,
            'DIDs': self._dids
        }

        all_system_info_scipy = {'AllSystemInfo': all_system_info}
        savemat(absolute_path, all_system_info_scipy, long_field_names=True, do_compression=True)
        self.info('%s - AllSystemInfo placed at %s', __name__, absolute_path)

    def get_core_ids(self):
        """Get core IDs for specified project.

        Returns:
            dict: DID as a dict with the following format
            {
                "unit name" : {}
            }

        """
        result = {}
        unit_configs = self.unit_cfg.get_per_unit_cfg()
        for unit, configs in unit_configs.items():
            result[unit] = deepcopy(configs['core'])
        for unit, configs in result.items():
            for core_id, core_id_dict in configs.items():
                for identifier, data_dict in core_id_dict.items():
                    for tl_field, tl_data in data_dict.items():
                        # Turn lists into numpy.array, required to get matlab cell arrays instead of char arrays.
                        # char arrays make DocGen crash.
                        if isinstance(tl_data, list):
                            configs[core_id][identifier][tl_field] = array(tl_data, dtype=object)
                        else:
                            configs[core_id][identifier][tl_field] = tl_data
        return result

    def get_dids(self):
        """Get DIDs for specified project.

        Returns:
            dict: DID as a dict with the following format
            {
                "unit name" : {}
            }

        """
        result = {}
        unit_configs = self.unit_cfg.get_per_unit_cfg()
        for unit, configs in unit_configs.items():
            dids_list = []
            for signal_values in configs['dids'].values():
                signal_values_allsysteminfo_keys = {
                        'sig_name': signal_values['name'],
                        'sig_desc': signal_values['description'],
                        'blk_name': signal_values['handle'],
                        'data_type': signal_values['type'],
                        'unit': signal_values['unit'],
                        'lsb': signal_values['lsb'],
                        'offset': signal_values['offset']
                        }
                dids_list.append(signal_values_allsysteminfo_keys)
            result.update({unit: dids_list})

        return result

    def _gen_dummy_file_info(self):
        """Generate dummy FileInfo struct for compatibility purposes."""
        result = {}
        for key in self._core_ids:
            result.update({key: ''})
        return result

    def get_signals_without_source(self):
        """Get missing and unused signals from project configuration.

        Returns:
            dict: result with the following format
            {
                "signal_name" : "signal_type"
            }

        """
        prj_conf = self.signal_if.check_config()
        result = self._get_external_signals(prj_conf)
        temp_signals = self._get_internal_signals(prj_conf)
        result = merge_dicts(result, temp_signals, merge_recursively=True)
        return result

    def _get_external_signals(self, prj_conf):
        signals_conf = prj_conf['sigs']['ext']
        return self._get_signals_by_types(signals_conf)

    def _get_internal_signals(self, prj_conf):
        result = {}
        signals_conf = prj_conf['sigs']['int']
        for sig_conf in signals_conf.values():
            temp_signals = self._get_signals_by_types(sig_conf)
            result = merge_dicts(result, temp_signals, merge_recursively=True)
        return result

    def _get_signals_by_types(self, signal_conf):
        """Get signals by multiple types and merge them into one dictionary, see _get_signals_by_type() for docs."""
        result = {}
        for signal_type in self._signal_types:
            temp_signals = _get_signals_by_type(signal_conf, signal_type)
            if temp_signals:
                result = merge_dicts(result, temp_signals, merge_recursively=True)
        return result



# Copyright 2024 Volvo Car Corporation
# Licensed under Apache 2.0.

"""Signal interface module.

Module for parsing the external signal interface definition files, and
checking the signal interfaces between Supplier and VCC sw,
and also internally between VCC sw-units.
"""
import re
import csv
import time
import copy
from pprint import pformat
from pathlib import Path
from collections import defaultdict

from powertrain_build.build_proj_config import BuildProjConfig
from powertrain_build.feature_configs import FeatureConfigs
from powertrain_build.lib.helper_functions import deep_dict_update
from powertrain_build.problem_logger import ProblemLogger
from powertrain_build.types import get_ec_type
from powertrain_build.unit_configs import UnitConfigs, CodeGenerators
from powertrain_build.user_defined_types import UserDefinedTypes


class CsvReaderCounter:
    """Csv wrapper to keep track of file row."""

    def __init__(self, _csvreader):
        """Init."""
        self.csvreader = _csvreader
        self.counter = 0

    def __iter__(self):
        """Iterate."""
        return self

    def __next__(self):
        """Next."""
        self.counter += 1
        return next(self.csvreader)

    def get_count(self):
        """Get current row number."""
        return self.counter


class InterfaceValueException(Exception):
    """Exception for errors interface list."""

    def __init__(self, signal_name, erroneous_field):
        """Init."""
        self.erroneous_field = erroneous_field
        self.signal_name = signal_name


class SignalInterfaces(ProblemLogger):
    """Base class for signal interfaces."""

    def __init__(self, unit_cfg, model_names=None):

        self._unit_cfg = unit_cfg
        self._signals_ext_nok = {'inports': {}, 'outports': {}}
        self._signals_ext_ok = {'inports': {}, 'outports': {}}
        self._result = self.__ddict_factory()

        if model_names is None:
            self.models_to_check = set()
        else:
            self.models_to_check = set(model_names)

    def get_external_outputs(self):
        """Get the external outputs.

        Should be implemented by the subclasses.

        Returns:
            dict: External outputs
        """
        raise NotImplementedError

    def get_external_inputs(self):
        """Get the internal outputs.

        Should be implemented by the subclasses.

        Returns:
            dict: Internal outputs
        """
        raise NotImplementedError

    @classmethod
    def __ddict_factory(cls):
        """Generate recursive defaultdict."""
        return defaultdict(cls.__ddict_factory)

    @property
    def should_all_models_be_checked(self):
        """True if there are no specific models to be checked

        :rtype: bool
        """
        return len(self.models_to_check) == 0

    def should_check_model(self, model_name):
        """True if MODEL_NAME is in SELF.MODEL_NAMES

        :rtype: bool
        """
        return self.should_all_models_be_checked or model_name in self.models_to_check

    def contains_model_to_check(self, model_names):
        """True if MODEL_NAMES contains a model in SELF.MODEL_NAMES

        :param model_names: iterable
        :rtype: bool
        """
        return any((m in self.models_to_check for m in model_names))

    @staticmethod
    def _eq_var(var_a, var_b, par):
        """Check if variables are equal, with some modifications.

        par contains the key for which type of parameter is compared
        E.g. nan is equal to '-' and ''. Furthermore, all variables
        are converted to strings before making the comparison.
        """
        a_s = str(var_a)
        b_s = str(var_b)
        if par in ['min', 'max']:
            a_s = re.sub('nan', '-', a_s)
            b_s = re.sub('nan', '-', b_s)
            if re.match(r'[-+.0-9eE]+', a_s) and re.match(r'[-+.0-9eE]+', b_s):
                try:
                    a_f = float(var_a)
                    b_f = float(var_b)
                    return a_f == b_f
                except ValueError:
                    pass
        return a_s == b_s

    def _check_var_def(self, ext_def, unit_def):
        """Check that the parameters for the variable definitions are the same.

        Check that the parameters are the same for all units that use the parameter.
        """
        res = {}
        ignored_parameters = ['class', 'configs', 'handle', 'default', 'description', 'unit']
        for unit, unit_definitions in unit_def.items():
            for parameter, definition in ext_def.items():
                if parameter in ignored_parameters or parameter not in unit_definitions:
                    continue
                unit_definition = unit_definitions[parameter]
                if not self._eq_var(definition, unit_definition, parameter):
                    res.setdefault(unit, {})[parameter] = f"{definition} != {unit_definition}"

        return res if res else None

    def _add_incons_to_ext_dict(self, res, var, incons):
        """Add inconsistencies to the ext dict."""
        if incons is not None:
            for unit, incon in incons.items():
                if self.should_check_model(unit):
                    if var in res['inconsistent_defs'][unit]:
                        res['inconsistent_defs'][unit][var].update(incon)
                    else:
                        res['inconsistent_defs'][unit][var] = incon

    def _add_incons_to_dict(self, res, var, incons):
        """Add iconsistent data to a dict."""
        if incons is not None:
            for unit, incon in incons.items():
                if self.should_check_model(unit):
                    if var in res[unit]['inconsistent_defs']:
                        res[unit]['inconsistent_defs'][var].update(incon)
                    else:
                        res[unit]['inconsistent_defs'][var] = incon

    def _gen_int_outp_set(self, tot_unit_cfg):
        """Generate a set of signals to check for internal output parameters."""
        tot_cfg = tot_unit_cfg.get('outports', {})
        external_ports = {**self._signals_ext_ok['outports'], **self._signals_ext_nok['outports']}

        # only check outports that are not defined in the external io definition
        non_ext_outp = set(tot_cfg.keys()) - set(external_ports.keys())
        return non_ext_outp

    def _gen_int_inp_set(self, tot_unit_cfg):
        """Generate a set of signals to check for internal input parameters."""
        tot_cfg = tot_unit_cfg.get('inports', {})
        external_ports = {**self._signals_ext_ok['inports'], **self._signals_ext_nok['inports']}

        # only check inports that are not defined in the external io definition
        non_ext_inp = set(tot_cfg.keys()) - set(external_ports.keys())
        return non_ext_inp

    def _gen_unit_var_dict(self, res, chk_type, variables, in_out):
        """Generate a dict with vars as keys, and units in which they are used as a list."""
        tot_cfg = self._unit_cfg.get_per_cfg_unit_cfg()
        for var in variables:
            for unit in tot_cfg[in_out][var]:
                if self.should_check_model(unit):
                    if var in res[unit][chk_type]:
                        res[unit][chk_type][var] = {}
                    else:
                        res[unit][chk_type][var] = {}

    def _check_external_outp(self):
        """Check that all external outputs (VCC -> supplier).

        Check that all external outputs are produced,
        that consumed signals (Supplier -> VCC) are defined as inputs
        in the external IO definition file, and that the signal
        definitions are consistent between io and unit definitions.
        """
        ext_out = self.get_external_outputs()
        tot_cfg = self._unit_cfg.get_per_cfg_unit_cfg()
        res = self._result['sigs']['ext']
        # check that all output signals are produced by VCC SW
        if 'outports' in tot_cfg:
            for _, data in ext_out.items():
                for var, var_def in data.items():
                    if var not in tot_cfg['outports']:
                        res['missing'][var] = {}
                        self._signals_ext_nok['outports'][var] = var_def
                    else:
                        self._signals_ext_ok['outports'][var] = var_def
                        tmp = self._check_var_def(var_def, tot_cfg['outports'][var])
                        self._add_incons_to_ext_dict(res, var, tmp)

    def _check_external_inp(self):
        """Check all external input signals.

        Check that outputs are produced and,
        that consumed signals are defined as inputs
        in the external IO definition file.
        """
        ext_inp = self.get_external_inputs()
        # self.debug('ext_inp: %s', pformat(ext_inp))
        tot_cfg = self._unit_cfg.get_per_cfg_unit_cfg()
        res = self._result['sigs']['ext']
        # check that all input signals are used by VCC SW
        if 'inports' in tot_cfg:
            for data in ext_inp.values():
                for var, var_def in data.items():
                    if var not in tot_cfg['inports']:
                        if var not in res['unused']:
                            res['unused'][var] = {}
                        self._signals_ext_nok['inports'][var] = var_def
                    else:
                        self._signals_ext_ok['inports'][var] = var_def
                        tmp = self._check_var_def(var_def, tot_cfg['inports'][var])
                        self._add_incons_to_ext_dict(res, var, tmp)

    def _check_internal_io(self):
        """Check internal signal io.

        Function which checks that:
        1. all signals consumed in models, are produced
        in antoher model.
        2. all outputs are consumed (warning)
        3. check that all signal definitions are the same as the producing unit.
        4. a signal is only produced in one model (per prj).
        this function return a tuple with the above content

        TODO:
            Shall we add a check for multiple inputs of the same signal?*
        """
        tot_cfg = self._unit_cfg.get_per_cfg_unit_cfg()
        internal_inport_signals = self._gen_int_inp_set(tot_cfg)
        internal_outport_signals = self._gen_int_outp_set(tot_cfg)
        all_outports = set(tot_cfg.get('outports', {}).keys())
        res = self._result['sigs']['int']
        # 1. all signals consumed in models, are produced in another model.
        missing = internal_inport_signals - all_outports
        self._gen_unit_var_dict(res, 'missing', missing, 'inports')  # error!
        # 2. all outputs are consumed (warning)
        unused = internal_outport_signals - internal_inport_signals
        self._gen_unit_var_dict(res, 'unused', unused, 'outports')  # warning?
        # 3. check that all signal definitions are the same as the producing
        # unit.
        input_signals_to_check = internal_inport_signals - missing
        for signal in input_signals_to_check:
            unit_key = list(tot_cfg['outports'][signal].keys())
            # if output is defined in multiple functions, test #4 will catch this
            if len(unit_key) == 1:
                outport_def = tot_cfg['outports'][signal][unit_key[0]]
                tmp = self._check_var_def(outport_def,
                                          tot_cfg['inports'][signal])
                self._add_incons_to_dict(res, signal, tmp)

        # 4. check that we have not screwed up any of the consuming models
        external_signals_to_check = internal_outport_signals - unused
        for signal in external_signals_to_check:
            unit_keys = list(tot_cfg['inports'][signal].keys())
            my_outport_definition = tot_cfg['outports'][signal]
            for unit_key in unit_keys:
                others_inport_def = tot_cfg['inports'][signal][unit_key]
                tmp = self._check_var_def(others_inport_def,
                                          my_outport_definition)
                self._add_incons_to_dict(res, signal, tmp)

        # 5. there should only be one outport per config.
        multiple_defs = []
        if 'outports' in tot_cfg:
            multiple_defs = [x for x in tot_cfg['outports']
                             if len(tot_cfg['outports'][x]) > 1]
        self._gen_unit_var_dict(res, 'multiple_defs', multiple_defs, 'outports')

    def _check_config(self):
        """Check the interfaces given a certain project config.

        The result of the checks is stored in the dict self._result
        """
        # reset the handled external signals for each configuration
        self._signals_ext_ok = {'inports': {}, 'outports': {}}
        # has to be done in this order since there is an implicit dependency
        # to self._signals_ext_ok
        self._check_external_outp()
        self._check_external_inp()
        self._check_internal_io()

    def check_config(self):
        """Check configurations for specific project.

        Returns:
            dict: the result of the check with the following format

        ::

            {
                "sigs": { "ext": {"missing": {},
                                  "unused": {},
                                  "inconsistent_defs": {}},
                          "int": {"UNIT_NAME": {"missing": {},
                                                "unused": {},
                                                "multiple_defs": {}
                                                "inconsistent_defs": {}}
                }
            }

        """
        self._result = self.__ddict_factory()
        self._check_config()
        # self.debug("%s", pformat(self._result))
        # restore the current conofiguration

        return self._result


class YamlSignalInterfaces(SignalInterfaces):
    """Interface configurations defined in yaml files."""

    @staticmethod
    def from_config_file(project_config_path):
        """Create a YamlSignalInterfaces instance from a project config file.

        Args:
            project_config_path (str): path to the project config file

        Returns:
            YamlSignalInterfaces: instance of YamlSignalInterfaces
        """
        build_project_config = BuildProjConfig(project_config_path)
        feature_cfg = FeatureConfigs(build_project_config)
        unit_config = UnitConfigs(build_project_config, feature_cfg)
        user_defined_types = UserDefinedTypes(build_project_config, unit_config)

        return YamlSignalInterfaces(
            prj_cfg=build_project_config,
            unit_cfg=unit_config,
            feature_cfg=feature_cfg,
            user_defined_types=user_defined_types
        )

    def __init__(self, prj_cfg, unit_cfg, feature_cfg, user_defined_types, model_names=None):
        """Class initializer.

        Args:
            prj_cfg (BuildProjConfig): configures which units are active in the current project.
            unit_cfg (UnitConfigs): class instance containing all the unit configuration parameters.
            feature_cfg (FeatureConfig): Feature configs from SPM_Codeswitch_Setup.
            user_defined_types (UserDefinedTypes): Class holding user defined data types.
            model_names (set): models that should be included in the check, default is all models
        """
        super().__init__(unit_cfg, model_names=model_names)

        # Postpone imports to here to work on machines without PyYaml installed
        from powertrain_build.interface.base import filter_signals
        from powertrain_build.interface.application import Application, get_internal_domain
        from powertrain_build.interface.generation_utils import get_interface, get_method_interface
        from powertrain_build.interface.hal import HALA
        from powertrain_build.interface.device_proxy import DPAL
        from powertrain_build.interface.zone_controller import ZCAL
        from powertrain_build.interface.service import ServiceFramework

        app = Application()
        app.pybuild['build_cfg'] = prj_cfg
        app.name = app.pybuild['build_cfg'].name
        app.pybuild['feature_cfg'] = feature_cfg
        app.pybuild['unit_vars'] = unit_cfg.get_per_unit_cfg_total()
        app.pybuild['user_defined_types'] = user_defined_types

        translation_files = app.get_translation_files()

        ecu_supplier = prj_cfg.get_ecu_info()[0]
        self.zc_spec = {}
        self.hal_spec = {}
        self.dp_spec = {}
        self.sfw_spec = {}
        self.sa_spec = {}
        self.service_spec = {}
        self.mthd_spec = {}
        if prj_cfg.get_code_generation_config(item='generateYamlInterfaceFile'):
            zc_app = ZCAL(app)
            self.zc_spec = get_interface(app, zc_app)
            self.composition_spec = zc_app.composition_spec
        elif ecu_supplier == 'HI':
            hi_app = DPAL(app)
            self.dp_spec = get_interface(app, hi_app)
        else:
            hala = HALA(app)
            dp = DPAL(app)
            swf = ServiceFramework(app)
            hala.parse_definition(translation_files)
            self.dp_spec = get_interface(app, dp)
            self.sfw_spec = get_interface(app, swf)
            self.mthd_spec = get_method_interface(app)
            self.service_spec = self.get_availability(app, unit_cfg.code_generators)
            rasters = app.get_rasters()
            self.debug('Rasters: %s', rasters)
            internal = get_internal_domain(rasters)
            properties_from_json = [
                {"destination": "min", "source": "min", "default": "-"},
                {"destination": "max", "source": "max", "default": "-"},
                {"destination": "variable_type", "source": "type"},
                {"destination": "offset", "source": "offset", "default": "-"},
                {"destination": "factor", "source": "lsb", "default": 1},
                {"destination": "description", "source": "description"},
                {"destination": "unit", "source": "unit", "default": "-"},
            ]
            for raster in rasters:
                hala.name = raster.name
                hala.clear_signal_names()
                hala.add_signals(filter_signals(raster.insignals, internal), 'insignals', properties_from_json)
                hala.add_signals(raster.outsignals, 'outsignals', properties_from_json)
                self.debug('Current HALA: %s', hala)
                self.hal_spec[raster.name] = hala.to_dict()

    @staticmethod
    def get_availability(app, code_generators={CodeGenerators.target_link}):
        """Get the availability of services.

        Args:
            app (Application): Application instance
            code_generators (set): Code generators to include in the availability check

        Returns:
            dict: Availability of services
        """
        tl_type = 'Bool'
        variable_type = tl_type if CodeGenerators.target_link in code_generators else get_ec_type(tl_type)
        services = app.get_service_mapping()
        spec = {}
        for interface, service in services.items():
            camel_interface = ''.join(part.title() for part in interface.split('_'))
            spec[interface] = {
                'variable': f'yVcSfw_B_{camel_interface}IsAvailable',
                'variable_type': variable_type,
                'property_type': 'bool',
                'service': service,
                'default': 0,
                'length': 1,
                'property': 'inherent',
                'offset': '-',
                'factor': 1,
                'range': {'min': '0', 'max': '1'},
                'init': 0,
                'description': f'Availability of {interface} in {service}',
                'unit': '-',
                'group': None,
                'model': interface
            }
        return spec

    def get_externally_defined_ports(self):
        """Get ports defined by suppliers."""
        outports = []
        for raster in self.dp_spec.values():
            for spec in raster['consumer']:
                outports.append(spec['variable'])
        for raster in self.hal_spec.values():
            for spec in raster['consumer']:
                outports.append(spec['variable'])
        for raster in self.sfw_spec.values():
            for spec in raster['consumer']:
                outports.append(spec['variable'])
        for raster in self.sa_spec.values():
            for spec in raster['consumer']:
                outports.append(spec['variable'])
        for raster in self.zc_spec.values():
            for spec in raster['consumer']:
                outports.append(spec['variable'])
        return outports

    def get_external_io(self):
        """Get the variable definitions for the signal IO for a given config.

        Returns:
            dict: Variable definitions for the supplier in-/out-put signals for the configuration
        """
        def normalize_spec(spec):
            """ Convert Yaml spec to normal pybuild spec.

            Arguments:
                spec (dict): Yaml specification of signal
            Returns:
                spec (dict): Pybuild specification of signal
            """
            spec2 = copy.copy(spec)  # copy by value
            spec2['type'] = spec['variable_type']
            spec2['min'] = spec['range']['min']
            spec2['max'] = spec['range']['max']
            return spec2

        def set_spec(target):
            new_spec = normalize_spec(spec)
            if new_spec.get('debug'):
                debug[target][spec['variable']] = new_spec
            elif spec.get('dependability'):
                dependability[target][spec['variable']] = new_spec
            else:
                normal[target][spec['variable']] = new_spec

        normal = {}
        dependability = {}
        secure = {}  # Only supported by CSV interfaces
        debug = {}

        for field in ['input', 'output', 'status']:
            normal[field] = {}
            dependability[field] = {}
            debug[field] = {}

        spec_confs = [self.dp_spec, self.sfw_spec, self.sa_spec, self.hal_spec, self.zc_spec]
        for spec_conf in spec_confs:
            for raster in spec_conf.values():
                for spec in raster['consumer']:
                    set_spec("input")
                for spec in raster['producer']:
                    set_spec("output")
        for spec in self.service_spec.values():
            set_spec('status')
        for method_data in self.mthd_spec.values():
            for _, spec in method_data['ports']['out'].items():
                set_spec('input')
            for _, spec in method_data['ports']['in'].items():
                set_spec('output')
        return normal, dependability, secure, debug

    def get_external_signals(self, find_output=True):
        """Get the external signals.

        Args:
            find_output (bool): True if output signals should be returned, False if input signals should be returned

        Returns:
            dict: External signals
        """
        io_type = self.get_external_io()  # (normal, dependability, secure (unsupported), debug)
        directional_io = defaultdict(dict)
        for io in io_type:
            for signal_type, signals in io.items():
                if find_output and signal_type == 'output':
                    directional_io[signal_type].update(signals)
                elif not find_output and signal_type == 'input':
                    directional_io[signal_type].update(signals)
        return directional_io

    def get_external_outputs(self):
        return self.get_external_signals(True)

    def get_external_inputs(self):
        return self.get_external_signals(False)


class CsvSignalInterfaces(SignalInterfaces):
    """Interface configurations for all units and all configs.

    Provides methods for retrieving the currently
    used signal configurations of a unit.
    """

    convs = (('~=', '!='), ('~', ' not '), ('!', ' not '), (r'\&\&', ' and '),
             (r'\|\|', ' or '))
    # Common excel sheet definitions, adapted for CVS.
    VAR_COL = 0  # Variable name
    DATA_COL = 2  # Type, min, max, unit, comment, init
    PROJ_COL = 9  # Variable number of projects
    HEADER_ROW = 3  # which row the header info is found
    DATA_ROW = 5  # the row which data starts
    WS_NAMES = ['EMS-Output', 'EMS-Input', 'LIN-Output', 'LIN-Input',
                'CAN-Output', 'CAN-Input', 'Private CAN-Output',
                'Private CAN-Input']

    def __init__(self, prj_cfg, unit_cfg, models=None):
        """Class initializer.

        Args:
            prj_cfg (BuildProjConfig): configures which units are active in the
                                              current project
            unit_cfg (UnitConfigs): class instance containing all the unit
                                    configuration parameters
            models (str or list): Models to get interface for. default: 'all'

        """
        super().__init__(unit_cfg, model_names=models)
        self.interfaces = {}

        self.__prjs = {}
        self.__out_dict = {}
        self.__all_prjs = set()
        self.__prj2col = {}
        self.__prj_col_range = {}

        prj2index = {}
        self.name = prj_cfg.get_prj_config()
        file_path = prj_cfg.get_if_cfg_dir()
        self._prj_cfg = prj_cfg
        self._parse_io_cnfg(file_path)
        for k in self.__prj2col:
            self.__prj2col[k] = prj2index

    def _parse_io_cnfg(self, interface_directory):
        """Parse the CSV config files."""
        start_time = time.time()
        self.info('******************************************************')
        self.info('Start parsing SPM-Interface definition files')
        for ws_name in self.WS_NAMES:
            self.debug("read sheet %s start", ws_name)
            file_path = Path(interface_directory, ws_name + '.csv')
            try:
                with open(file_path, newline='') as fhandle:
                    if ws_name not in self.interfaces:
                        self.interfaces.update({ws_name: {}})
                    interface = self.interfaces[ws_name]
                    csvreader = csv.reader(fhandle, delimiter=';', strict=True)
                    csvreader = CsvReaderCounter(csvreader)
                    try:
                        for _ in range(self.HEADER_ROW-1):
                            next(csvreader)
                        row = next(csvreader)
                        # TODO: Find why excel export script sometimes skips first field
                        #       on each line if field is empty.
                        #       Workaround:
                        row = row[1:] if row[0] == '' else row
                        try:
                            col = self._get_proj_col(row)
                        except IndexError:
                            self.warning('Project %s not defined in %s', self.name, file_path.stem)
                            continue

                        for _ in range(self.HEADER_ROW, self.DATA_ROW-1):
                            next(csvreader)
                    except StopIteration:
                        self.critical('File %s has bad format, not enough header rows', file_path.stem)
                    for row in csvreader:
                        row = row[1:] if row[0] == '' else row
                        try:
                            signal, data = self._get_var_def(row)
                            data.update({'element_index': csvreader.get_count()})
                        except InterfaceValueException as interface_exception:
                            self.critical('Missing value for key "%s" in list "%s" at index "%s"',
                                          interface_exception.erroneous_field,
                                          file_path.stem, csvreader.get_count())
                            continue
                        if signal not in interface:
                            interface.update({signal: data})
                        interface[signal]['IOType'] = row[col].lower()
                self.debug("read sheet %s end", ws_name)
            except FileNotFoundError:
                self.info('Project %s does not have a %s', self.name, ws_name)
        self.info('Finished parsing SPM-Interface definition file (in %4.2f s)', time.time() - start_time)

    def __repr__(self):
        """Get string representation of object."""
        return pformat(self.__out_dict)

    def _get_proj_col(self, row):
        """Get the projects in the config document.

        Parses supplied worksheet row and finds all the projects defined, and
        stores the result in the class variable __all_prjs.
        Furthermore, a dict prj2col, maps the project name to a list index
        in the prj key in the __out_dict dict, and
        sets the class variables __prj2col, and __prj_col_range contains
        a list with column indexes in the excel sheet, used for parsing the
        excel, the result is stored in the interal __out_dict

        interfaces example:

        ::

            {
                'CAN-Input': {
                    'sVCcm_D_HvacCoolgEnaRe': {
                        'IOType': 'd'
                        'description': 'Enable Ac rear HVAC',
                        'init': 0,
                        'max': 1,
                        'min': 0,
                        'type': 'UInt8',
                        'unit': '-'
                    }
                }
            }

        """
        col = self.PROJ_COL
        while row[col] is not None and row[col] != '':
            if row[col] == self.name:
                return col
            col += 1
            if col > 200:
                break

        raise IndexError

    @staticmethod
    def _get_var_def(row):
        """Get the variable definition from a row in a sheet.

        Returns:
            dict: with var definitions

        """
        keys = ['type', 'min', 'max', 'unit', 'description', 'init']

        vals = []
        for col in [0, 2, 3, 4, 5, 6, 7]:
            val = row[col]
            if isinstance(val, str):
                val = val.strip()
                if re.fullmatch('[+-]?[0-9]+', val):
                    val = int(val)
                elif re.fullmatch('[+-]?[0-9.,]+', val):
                    val = float(val.replace(',', '.'))
            vals.append(val)
        critical_attributes = ['type']
        signal_name = vals[0]
        signal_attributes = dict(zip(keys, vals[1:]))
        for field in critical_attributes:
            if signal_attributes[field] == '':
                raise InterfaceValueException(signal_name, field)
        return signal_name, signal_attributes

    def get_raw_io_cnfg(self):
        """Get the raw IO-Config parsed from the config file.

        Returns:
            dict: raw configuration information from the config file

        dict example:

        ::

            {
                'CAN-Input': {
                    'sVCcm_D_HvacCoolgEnaRe': {
                        'description': 'Enable Ac rear HVAC',
                        'init': 0,
                        'max': 1,
                        'min': 0,
                        'project_def': 'D',
                        'type': 'UInt8',
                        'unit': '-'
                    }
                }
            }


        """
        return self.interfaces

    def get_io_config(self):
        """Get an IO-Config for a specific project.

        Returns:
            tuple: a tuple with three dicts - (nrm_dict, dep_dict, dbg_dict)

        """
        normal = {}
        dep = {}
        secure = {}
        debug = {}
        for interface_name, interface in self.interfaces.items():
            normal[interface_name] = {}
            dep[interface_name] = {}
            secure[interface_name] = {}
            debug[interface_name] = {}
            for signal_name, signal in interface.items():
                prj_def = signal.get('IOType', '-')
                if prj_def in 'xd':
                    normal[interface_name][signal_name] = copy.deepcopy(signal)
                elif prj_def in 's':
                    dep[interface_name][signal_name] = copy.deepcopy(signal)
                elif prj_def == 'sc':
                    secure[interface_name][signal_name] = copy.deepcopy(signal)
                if prj_def in 'd':
                    debug[interface_name][signal_name] = copy.deepcopy(signal)
        return normal, dep, secure, debug

    def get_external_io(self):
        """Get the variable definitions for the signal IO for a given config.

        Returns:
            dict: Variable definitions for the supplier in-/out-put signals for the configuration

        """
        return self.get_io_config()

    def get_external_inputs(self):
        """Get variable definitions for the supplier signal Inputs.

        Returns:
            dict: variable definitions for the supplier input signals

        """
        cfg, cfg_dep, cfg_sec = self.get_io_config()[:3]
        deep_dict_update(cfg, cfg_dep)
        deep_dict_update(cfg, cfg_sec)
        out = {}
        for key, value in cfg.items():
            if re.match('.*Input$', key) is not None:
                out[key] = value
        return out

    def get_external_outputs(self):
        """Get variable definitions for the supplier signal Outputs.

        Returns:
            dict: variable definitions for the supplier output signals

        """
        cfg, cfg_dep, cfg_sec = self.get_io_config()[:3]
        deep_dict_update(cfg, cfg_dep)
        deep_dict_update(cfg, cfg_sec)
        out = {}
        for key, value in cfg.items():
            if re.match(r'.*Output$', key) is not None:
                out[key] = value
        return out

    def get_externally_defined_ports(self):
        """Get ports defined by suppliers."""
        externally_defined_ports = set()
        external_outports = self.get_external_outputs()
        for external_ports in external_outports.values():
            externally_defined_ports.update(external_ports)
        external_inports = self.get_external_inputs()
        for external_ports in external_inports.values():
            externally_defined_ports.update(external_ports)
        return list(externally_defined_ports)



# Copyright 2024 Volvo Car Corporation
# Licensed under Apache 2.0.

# -*- coding: utf-8 -*-
"""Python module used for calculating interfaces for CSP"""
import argparse
import sys
from pathlib import Path
from os import path
from typing import List, Optional

from powertrain_build.interface.hal import HALA
from powertrain_build.interface.device_proxy import DPAL
from powertrain_build.interface.service import ServiceFramework
from powertrain_build.interface import simulink
from powertrain_build.lib import logger
from powertrain_build.interface import generation_utils
from powertrain_build.lib.helper_functions import deep_json_update

LOGGER = logger.create_logger("CSP adapters")

PARSER_HELP = "Generate adapters"


def configure_parser(parser: argparse.ArgumentParser):
    generation_utils.add_base_args(parser)
    parser.add_argument(
        "--dp-interface",
        help="Add dp interface to adapter specification",
        action="store_true"
    )
    parser.add_argument(
        "--hal-interface",
        help="Add dp interface to adapter specification",
        action="store_true"
    )
    parser.add_argument(
        "--service-interface",
        help="Add sfw interface to adapter specification",
        action="store_true"
    )
    parser.add_argument(
        "output",
        help="Output file with interface specifications",
        type=Path
    )
    parser.add_argument(
        "--update-config",
        help="Update project config file with path to adapter specifications",
        action="store_true"
    )
    parser.set_defaults(func=generate_adapters)


def main(argv: Optional[List[str]] = None):
    """ Main function for stand alone execution.
    Mostly useful for testing and generation of dummy hal specifications
    """
    parser = argparse.ArgumentParser(description=PARSER_HELP)
    configure_parser(parser)
    args = parser.parse_args(argv)
    args.func(args)


def generate_adapters(args: argparse.Namespace):
    app = generation_utils.process_app(args.config)
    adapters(args, app)


def update_project_config(args):
    """ Update project config file with relative location to adapter specification file, linux styled path.

    Args:
        args (Namespace): Arguments from command line
    """
    config_dir = Path(args.config).resolve().parents[0]
    output_dir = Path(args.output).resolve().parents[0]
    rel_dir = Path(path.relpath(output_dir, config_dir))
    rel_path = rel_dir / Path(args.output).name
    rel_path_linux = rel_path.as_posix()
    deep_json_update(
        args.config,
        {'ProjectInfo': {'adapterSpec': rel_path_linux}}
    )


def adapters(args, app):
    """ Generate specification for adapter generation.

    Args:
        args (Namespace): Arguments from command line
        app (Application): Application to generate specification for
    """

    if args.dp_interface:
        dp = DPAL(app)
        dp_interface = generation_utils.get_interface(app, dp)
    else:
        dp_interface = {}

    if args.hal_interface:
        hala = HALA(app)
        hal_interface = generation_utils.get_interface(app, hala)
    else:
        hal_interface = {}

    if args.service_interface:
        sfw = ServiceFramework(app)
        sfw_interface = generation_utils.get_interface(app, sfw)
        method_interface = generation_utils.get_method_interface(app)
    else:
        sfw_interface = {}
        method_interface = {}

    interface_data_types = app.pybuild['user_defined_types'].get_interface_data_types()

    adapters = simulink.get_interface(
        interface_data_types,
        dp_interface,
        hal_interface,
        sfw_interface,
        method_interface
    )

    generation_utils.write_to_file(adapters, args.output, is_yaml=True)

    if args.update_config:
        update_project_config(args)


if __name__ == "__main__":
    main(sys.argv[1:])
